# -*- coding: utf-8 -*-
"""
Created on Fri Jul 21 15:51:57 2023

@author: anba
"""
import numpy as np 
import sys
import copy
import rich
import pandas as pd
#if __name__ == '__main__':
# sys.path.insert(1, r'C:\Users\anba\Desktop\Projects\pyplume\pyplume\ptools')
# import ptools


from ..ptools import ptools


class geometry:
    def __init__(self,pd0,**kwargs):
        
        self._pd0 = pd0
        
        self.instrument_depth = 0 # instrument height above bed
        self.instrument_HAB = 0 # intrument height above bed
        
        
    def get_bin_midpoints(self):
        """
        Generate numpy array of bin midpoint distances in meters. 
        Args:
            None
        Returns:
            numpy array with dimensions (n_ensembles)  
        """          
        
        # if self._pd0.config.beam_facing == 'DOWN':
        #     direction = -1
        # else: direction = 1
        
        beam_length = self._pd0.config.number_of_cells*self._pd0.config.depth_cell_length # total length of the beam 
        
        
        bin_distances = np.linspace(self._pd0.config.bin_1_distance,
                                    beam_length+self._pd0.config.bin_1_distance,
                                    self._pd0.config.number_of_cells)/100
        
        
            
        
        
        return bin_distances 
    
    def get_bin_midpoints_depth(self):
        """
        Generate numpy array of bin midpoint depths, based on static instrument depth. 
        Args:
            None
        Returns:
            numpy array with dimensions (n_ensembles)  
        """   
        if self._pd0.config.beam_facing == 'DOWN':
            bin_midpoints_depth = self.instrument_depth + self.get_bin_midpoints() 
        elif self._pd0.config.beam_facing =='UP':
            bin_midpoints_depth = self.instrument_depth - self.get_bin_midpoints() 
        return bin_midpoints_depth
    
    def get_bin_midpoints_HAB(self):
        """
        Generate numpy array of bin midpoint heights, based on static instrument heihgt above bed. 
        Args:
            None
        Returns:
            numpy array with dimensions (n_ensembles)  
        """  
        if self._pd0.config.beam_facing == 'DOWN':
            bin_midpoints_HAB = self.instrument_HAB - self.get_bin_midpoints()
        elif self._pd0.config.beam_facing =='UP':
            bin_midpoints_HAB = self.instrument_HAB + self.get_bin_midpoints()
        return bin_midpoints_HAB         
    
    
    
    # def get_absolute_beam_midpoint_positions(self):
    #     """
    #     Calculate the absolute position of each beam, taking into account the 
    #     position and orientation (pitch,roll,heading) of the platform (e.g., ROV), 
    #     and the relative geometry of the beam (e.g.,beam angle, bin size) 
 
 
    #     Returns
    #     -------
    #     absolute_beam_midpoint_positions: list of numpy arrays
    #         list of beam midpoint positions for each beam and each ensemble.  
    #         Element of the list contains the array for each beam, and each array
    #         has dimensions of (3,n_bins,n_ensembles)
            
 
    #     """
        
    #     ## get position of beam midpoints for every ensemble 
    #     absolute_beam_midpoint_positions = []
    #     X = np.empty((3,self._pd0.config.number_of_beams,self._pd0.config.number_of_cells,self._pd0.n_ensembles))
    #     pose = self.pose.pose[['x','y','z']].to_numpy().T #position  
    #     for b in range(self._pd0.config.number_of_beams):
    #         beam = self.relative_beam_midpoint_positions[b]
    #         #absolute_beam_midpoint_positions.append(np.full((3,self._pd0.config.number_of_cells,self._pd0.n_ensembles),0, dtype = float))
    #         for e in range(self._pd0.n_ensembles):
                
    #             #build rotation matrix
                
         
    #             yaw = self.pose.pose['heading'].to_numpy()[e]
    #             pitch = self.pose.pose['pitch'].to_numpy()[e]
    #             roll = self.pose.pose['roll'].to_numpy()[e]
                
    
                
    #             # yaw   = self.orientation[2][e]
    #             # pitch = self.orientation[0][e]
    #             # roll  = self.orientation[1][e]
    #             R = np.dot(ptools.gen_rot_x(roll),ptools.gen_rot_z(yaw).dot(ptools.gen_rot_y(pitch)))
    #             Xp = np.add(pose[:,e],beam.T.dot(R)).T
    #             X[:,b,:,e] = Xp#np.add(pos[:,e],beam.T).T        
    #     #% apply masks 
    #     # X = np.empty((self-mask(.config.number_of_beams,self.config.number_of_cells,self.n_ensembles))
        
        
    #     X[0] = self._pd0.mask.apply_masks(X[0])
    #     X[1] = self._pd0.mask.apply_masks(X[1])
    #     X[2] = self._pd0.mask.apply_masks(X[2])
        
    #     return X    
    
   
        

        
############ Set Functions ####################################################



    def set_pose(self,Pose,update_orientation = False,**kwargs):
        """
        Define the instrument pose (position and orientation) with a 
        Pose object. Makes a cope of the pose object and stores it under 
        ctd.geometry.Pose
        
        Parameters
        ----------
        Pose : Pose
            Pose object containing position and orientation information.
        update_orientation : bool, optional 
            If True, the adcp orientation (from internal gyroscope) will be 
            used as the pose orientation (pitch, roll, and heading/yaw)
    
        method : 'nearest', 'ffill','bfill'
        
        **kwargs to send to pd.DataFrame.reindex
        Returns
        -------
        None
        """
        # copy input object and Resample the input timeseries to the ADCP ensemble timesteps
        self.pose = copy.deepcopy(Pose)
        
        

        self.pose.resample_to(self._pd0.get_ensemble_datetimes(),kwargs)
        
        if update_orientation:
            self.pose.pose['pitch'] = self._pd0.get_sensor_pitch()
            self.pose.pose['roll'] = self._pd0.get_sensor_roll()
            self.pose.pose['heading'] = self._pd0.get_sensor_heading()
            
            
        
            
        
        
        
        
        
        
        
        
        
        
    # def set_sensor_orientation(self,**kwargs):
    #     """
    #     Set the orientation data for the sensor. Accepts input orientation
    #     timeseries, and assigns instrument position based on the closest position 
    #     timestamp to each ensemble timestamp. All orientations are in degrees, 
    #     with pitch and roll ranging[-180,179.99],and heading ranging from [0,359.99] 
        
        
    
    #     Parameters
    #     ----------
    #     orient_in : (optional) pandas dataframe containing a timeseries of roll,pitch,heading 
    #             data corresponding to the orientation of the instrument. Must have 
    #             columns named 'pitch', 'roll' and 'heading'. Index must be a pandas 
    #             timestamp. If not specified, then the onboard gyroscope and compass data 
    #             (from the variable headers) will be used. 
    
    #     Returns
    #     -------
    #     None.
    
    #     """
        
    #     self.orientation = np.full((3,self._pd0.n_ensembles),np.nan)
        
        
    #     orient_in = kwargs.pop('orient_in',None)
        
    #     #if kwargs.get('orient_in'):
    #     if type(orient_in) == type(None):
    #         print('using sensor data')
    #         self.orientation[0,:] = self._pd0.get_sensor_pitch()
    #         self.orientation[1,:] = self._pd0.get_sensor_roll()
    #         self.orientation[2,:] = self._pd0.get_sensor_heading()
    #     else:
    #         et = self._pd0.get_ensemble_datetimes() # ensemble times
    #         for i,etime in enumerate(et):
    #             index = orient_in.index.get_indexer([etime], method='ffill')[0] # index of nearest input positon to the ensemble timestamp
    #             self.orientation[0,i] = orient_in.iloc[index].pitch
    #             self.orientation[1,i] = orient_in.iloc[index].roll
    #             self.orientation[2,i] = orient_in.iloc[index].heading 

            
        
    # def set_sensor_position(self,pos_in):
    #     """
    #     Set the position data for the sensor. Accepts input position timeseries,
    #     and assigns instrument position based on the closest position timestamp to
    #     each ensemble timestamp. 
    
    #     Parameters
    #     ----------
    #     pos_in : pandas dataframe
    #         containing a timeseries of x,y,z data corresponding to the position of 
    #         the instrument. Must have columns named 'x', 'y' and 'z'. Index must be
    #         a pandas timestamp. 
            
    
    #     Returns
    #     -------
    #     None.
    
    #     """
    #     et = self._pd0.get_ensemble_datetimes() # ensemble times
    #     self.position  = np.full((3,self._pd0.n_ensembles),np.nan)
    #     for i,etime in enumerate(et):
    #         index = pos_in.index.get_indexer([etime], method='ffill')[0] # index of nearest input positon to the ensembel timestamp

    #         self.position[0,i] = pos_in.iloc[index].x
    #         self.position[1,i] = pos_in.iloc[index].y
    #         self.position[2,i] = pos_in.iloc[index].z     
        
                    
    def calculate_beam_geometry(self,rotation = 0,offset = (0,0,0), dr = 0.1):
        """
    
        calculate the relative beam midpoint position for every ensemble (accounting for sensor orientation and platform orientation)
        calculate the true bottom track distance (accounting for orientation). 
        calculate the absolute position of each bin/ensemble in depth and relative to bottom track
        
        ----------
        rotation : float, optional
            Orientation of the self._pd0. This is the number of degrees clockwise 
            from the forward direction of the mounting platform. (E.g., the bow
            of the vessel). The default is 0.
        offset : numpy array, optional
            x,y,z offsets for the instrument. This offset 
            should be the relative distance between the position measuremet
            (e.g., from the vessel GPS) and the center of the ADCP transducer 
            face.The default is (0,0,0).
        dr : float, optional
            radial distance from the center of the face of the ADCP to the center 
            of the transdecer faces. The default is 0.1.
    
        Returns
        -------
        None.
    
        """
        
        # offset = (0,0,0)
        # dr = 0.1
        # rotation = 0
        ## calculate the orientation of the beam origins (e.g., orientation on face of ADCP) including the rotation of the insttuymet. (rotation of 0 means beam 3 is pointing "forwards" - in the positive x direction)
        R = ptools.gen_rot_z(theta = rotation)
        if self._pd0.config.beam_facing == 'DOWN':
            relative_beam_origin = np.array([(dr,0,0),(-dr,0,0),(0,dr,0),(0,-dr,0)])# position of each beam relative to the center of the face of the instrument
        if self._pd0.config.beam_facing == 'UP':
            relative_beam_origin = np.array([(-dr,0,0),(dr,0,0),(0,dr,0),(0,-dr,0)])# position of each beam relative to the center of the face of the instrument
        
        relative_beam_origin = np.add(offset,relative_beam_origin)  # offset origin to be relative to the center of mass of the position measurement
        relative_beam_origin = relative_beam_origin.dot(R).T # rotate the origin points - assums that positive y direction is forward (beam 3) 
        
        #%
        #% build an array containing the unrotated beam vectors (at this point they will all be pointing directly downward)
        relative_beam_midpoint_positions = np.full((3,self._pd0.config.number_of_beams,self._pd0.config.number_of_cells,self._pd0.n_ensembles),0, dtype = float)
        
        
        #X = np.empty((3,self._pd0.config.number_of_beams,self._pd0.config.number_of_cells,self._pd0.n_ensembles))
        X = np.repeat(self.pose.pose[['x','y','z']].to_numpy().T[:,np.newaxis,:],self._pd0.config.number_of_cells, axis = 1)
        X = np.repeat(X[:,np.newaxis,:,:], self._pd0.config.number_of_beams,axis = 1)
        X_hab = X.copy()
        
        
        
       
        bt_vec = np.full((3,self._pd0.config.number_of_beams,self._pd0.n_ensembles),0, dtype = float)
        
        bt_range = self._pd0.get_bottom_track()
        if any(~np.isnan(bt_range).flatten()):
            # bottom track range for each beam and ensemble 
            bt_vec[2] = -bt_range
    
        for b in range(self._pd0.config.number_of_beams):
            
            
            # array holding beam midpoints coordinates 
            beam_midpoints = np.repeat(np.outer((0,0,0),np.ones(self._pd0.config.number_of_cells))[:, :, np.newaxis], self._pd0.n_ensembles, axis=2)
            
            ## overwrite z-offset coordiate with HAB(measurement doesnt care about sensor z-offset at origin)
            #beam_midpoints[2] = Z[beam_no]
            
            if self._pd0.config.beam_facing == 'DOWN':
                beam_midpoints[2] += -np.repeat(self.get_bin_midpoints()[:, np.newaxis], self._pd0.n_ensembles, axis=1) #+ 1.18
            else: 
                beam_midpoints[2] += np.repeat(self.get_bin_midpoints()[:, np.newaxis], self._pd0.n_ensembles, axis=1)
            
            # #% rotation matricies for 
            theta_beam = self._pd0.config.beam_angle
            Ry_cw = ptools.gen_rot_y(-theta_beam)
            Rx_cw = ptools.gen_rot_x(-theta_beam)
            Ry_ccw = ptools.gen_rot_y(theta_beam)
            Rx_ccw = ptools.gen_rot_x(theta_beam)   
            
            #% rotate the bottom track vectors 
            
            #% account for beam geometry
            for e in range(self._pd0.n_ensembles):
                if b == 0:
                    relative_beam_midpoint_positions[:,0,:,e] = Ry_cw.dot(beam_midpoints[:,:,e])
                    bt_vec[:,b,e] = Ry_cw.dot(bt_vec[:,b,e])
                elif b == 1:
                    relative_beam_midpoint_positions[:,1,:,e] = Ry_ccw.dot(beam_midpoints[:,:,e])
                    bt_vec[:,b,e] = Ry_ccw.dot(bt_vec[:,b,e])
                elif b == 2:
                    relative_beam_midpoint_positions[:,2,:,e] = Rx_ccw.dot(beam_midpoints[:,:,e])
                    bt_vec[:,b,e] = Rx_ccw.dot(bt_vec[:,b,e])
                elif b == 3: 
                    relative_beam_midpoint_positions[:,3,:,e] = Rx_cw.dot(beam_midpoints[:,:,e])  
                    bt_vec[:,b,e] = Rx_cw.dot(bt_vec[:,b,e])
                   
                yaw = self.pose.pose['heading'].to_numpy()[e]
                pitch = self.pose.pose['pitch'].to_numpy()[e]
                roll = self.pose.pose['roll'].to_numpy()[e]
                
                R = np.dot(ptools.gen_rot_x(roll),ptools.gen_rot_z(yaw).dot(ptools.gen_rot_y(pitch)))
                relative_beam_midpoint_positions[:,b,:,e] = relative_beam_midpoint_positions[:,b,:,e].T.dot(R).T
 
                bt_vec[:,b,e] = bt_vec[:,b,e].dot(R).T
                
                
                #Xp =np.repeat(self.pose.pose[['x','y','z']].to_numpy().T[:,e])[:, np.newaxis], self._pd0.n_ensembles, axis=1) relative_beam_midpoint_positions[:,b,:,e]
                
                X[:,b,:,e] +=relative_beam_midpoint_positions[:,b,:,e]
                  
                X_hab[:,b,:,e] += relative_beam_midpoint_positions[:,b,:,e]
                X_hab[2,b,:,e] = relative_beam_midpoint_positions[2,b,:,e] - bt_vec[2,b,e] 
                
                
        self.relative_beam_midpoint_positions = relative_beam_midpoint_positions
        self.corrected_bottom_track = bt_vec[2]
        
        instrument_range = (self._pd0.config.bin_1_distance + self._pd0.config.depth_cell_length*self._pd0.config.number_of_cells)/100
        #self.corrected_bottom_track[self.corrected_bottom_track>instrument_range] = np.nan #reject bad data, replace with unrotated values 
        
        self.corrected_bottom_track[self.corrected_bottom_track>instrument_range] = bt_range[self.corrected_bottom_track>instrument_range]
        
        
        self.absolute_beam_midpoint_positions = X
        self.absolute_beam_midpoint_positions_HAB = X_hab
        #return X,X_hab,bt_vec[2]
        
    def get_relative_beam_midpoint_positions(self, mask = True):
        """
        retrieve absolute x,y,z position of each bin and ensemble with masks applied. 
 
 
        Returns
        -------
        absolute_beam_midpoint_positions: list of numpy arrays
            list of beam midpoint positions for each beam and each ensemble.  
            Element of the list contains the array for each beam, and each array
            has dimensions of (3,n_bins,n_ensembles)
            
 
        """
        X = self.relative_beam_midpoint_positions
        if mask:
            X[0] = self._pd0.mask.apply_masks(X[0])
            X[1] = self._pd0.mask.apply_masks(X[1])
            X[2] = self._pd0.mask.apply_masks(X[2])
            
        return X 
        
    def get_absolute_beam_midpoint_positions(self, mask = True):
        """
        retrieve absolute x,y,z position of each bin and ensemble with masks applied. 
 
 
        Returns
        -------
        absolute_beam_midpoint_positions: list of numpy arrays
            list of beam midpoint positions for each beam and each ensemble.  
            Element of the list contains the array for each beam, and each array
            has dimensions of (3,n_bins,n_ensembles)
            
 
        """
        X = self.absolute_beam_midpoint_positions
        if mask:
            X[0] = self._pd0.mask.apply_masks(X[0])
            X[1] = self._pd0.mask.apply_masks(X[1])
            X[2] = self._pd0.mask.apply_masks(X[2])
            
        return X 
    
    def get_absolute_beam_midpoint_positions_HAB(self, mask = True):
        """
        retrieve absolute x,y,z position of each bin and ensemble with masks applied. 
 
 
        Returns
        -------
        absolute_beam_midpoint_positions: list of numpy arrays
            list of beam midpoint positions for each beam and each ensemble.  
            Element of the list contains the array for each beam, and each array
            has dimensions of (3,n_bins,n_ensembles)
            
 
        """
        X = self.absolute_beam_midpoint_positions_HAB
        if mask:
            X[0] = self._pd0.mask.apply_masks(X[0])
            X[1] = self._pd0.mask.apply_masks(X[1])
            X[2] = self._pd0.mask.apply_masks(X[2])
            
        return X 
    def get_corrected_bottom_track(self):
        """
        retrieve absolute x,y,z position of each bin and ensemble with masks applied. 
 
 
        Returns
        -------
        absolute_beam_midpoint_positions: list of numpy arrays
            list of beam midpoint positions for each beam and each ensemble.  
            Element of the list contains the array for each beam, and each array
            has dimensions of (3,n_bins,n_ensembles)
            
 
        """
        X = self.corrected_bottom_track
            
        return X 
    
    def set_beam_relative_geometry(self,rotation = 0,offset = (0,0,0), dr = 0.1):
        """
        Calculate the x,y,z position of each bin in the coordinate reference 
        frame of either the the position measurement for the instrument. Accounts 
        for instrument beam angle (e.g., 20 deg) in x-z and y-z planes. Positions are in the reference
        frame of the instrument position measurement.
        Parameters
        ----------
        rotation : float, optional
            Orientation of the ADCP. This is the number of degrees clockwise 
            from the forward direction of the mounting platform. (E.g., the bow
            of the vessel). The default is 0.
        offset : numpy array, optional
            x,y,z offsets for the instrument. This offset 
            should be the relative distance between the position measuremet
            (e.g., from the vessel GPS) and the center of the ADCP transducer 
            face.The default is (0,0,0).
        dr : float, optional
            radial distance from the center of the face of the ADCP to the center 
            of the transdecer faces. The default is 0.1.

        Returns
        -------
        None.

        """

    
        # See WorkHorseCommands and Ouput Data Format page 53
        # rotation = 45
        # offset = adcp_offset
        
        # dr = 0.1 # distance from center of the unit to the center of each transducer face
        
        
        ## calculate the orientation of the beam origins (e.g., orientation on face of ADCP) including the rotation of the insttuymet. (rotation of 0 means beam 3 is pointing "forwards" - in the positive x direction)
        R = ptools.gen_rot_z(theta = rotation)
        if self._pd0.config.beam_facing == 'DOWN':
            self.relative_beam_origin = np.array([(dr,0,0),(-dr,0,0),(0,dr,0),(0,-dr,0)])# position of each beam relative to the center of the face of the instrument
        if self._pd0.config.beam_facing == 'UP':
            self.relative_beam_origin = np.array([(-dr,0,0),(dr,0,0),(0,dr,0),(0,-dr,0)])# position of each beam relative to the center of the face of the instrument
            
        self.relative_beam_origin = np.add(offset,self.relative_beam_origin)  # offset origin to be relative to the center of mass of the position measurement
        self.relative_beam_origin = self.relative_beam_origin.dot(R).T # rotate the origin points - assums that positive y direction is forward (beam 3) 
        
        
        
        #% build an array containing the unrotated beam vectors (at this point they will all be pointing directly downward)
        self.relative_beam_midpoint_positions = []
        for b in range(self._pd0.config.number_of_beams):
            origin = self.relative_beam_origin[:,b]
            beam_midpoints = np.outer(origin,np.ones(self._pd0.config.number_of_cells))
            if self._pd0.config.beam_facing == 'DOWN':
                beam_midpoints[2] += -self.get_bin_midpoints()
            else: 
                beam_midpoints[2] += self.get_bin_midpoints()
            self.relative_beam_midpoint_positions.append(beam_midpoints)
        
        
        # rotate each beam about its origin point by theta_beam. 
        
        #theta_beam = self._pd0.ensemble_data[0]['FIXED LEADER']['BEAM ANGLE']
        theta_beam = self.config.beam_angle
 
        Ry_cw = ptools.gen_rot_y(-theta_beam) # rotate clockwise about the y-axis
        Rx_cw = ptools.gen_rot_x(-theta_beam) # rotate clockwise about the x-axis
        Ry_ccw = ptools.gen_rot_y(theta_beam) # rotate counter-clockwise about the y-axis
        Rx_ccw = ptools.gen_rot_x(theta_beam) # rotate counter-clockwise about the y-axis
        
        # self.relative_beam_midpoint_positions[0] = Rx_cw.dot(Ry_cw.dot(self.relative_beam_midpoint_positions[0]))
        # self.relative_beam_midpoint_positions[1] = Rx_ccw.dot(Ry_ccw.dot(self.relative_beam_midpoint_positions[1]))
        # self.relative_beam_midpoint_positions[2] = Rx_ccw.dot(Ry_cw.dot(self.relative_beam_midpoint_positions[2]))
        # self.relative_beam_midpoint_positions[3] = Rx_cw.dot(Ry_ccw.dot(self.relative_beam_midpoint_positions[3]))     
        
        self.relative_beam_midpoint_positions[0] = Ry_cw.dot(self.relative_beam_midpoint_positions[0])
        self.relative_beam_midpoint_positions[1] = Ry_ccw.dot(self.relative_beam_midpoint_positions[1])
        self.relative_beam_midpoint_positions[2] = Rx_ccw.dot(self.relative_beam_midpoint_positions[2])
        self.relative_beam_midpoint_positions[3] = Rx_cw.dot(self.relative_beam_midpoint_positions[3])  
        
        
        
        



# -*- coding: utf-8 -*-
"""
Created on Wed Jul 19 15:24:17 2023

@author: anba
"""

import struct,os,sys,time,datetime


import numpy as np
import re

#sys.path.insert(1, r'C:\Users\anba\Desktop\Projects\pyplume\pyplume\ptools')
#import ptools


# # # from plotting import plotting
# # # from processing import processing
# # # from geometry import geometry
# # # from masking import masking
# # # from PT3 import PT3


from ..ptools import ptools
from .plotting import plotting
from .processing import processing
from .geometry import geometry
from .masking import masking
from .PT3 import PT3



#%% Data formats from 

# WorkHorse Commands and Output Data Format

ensemble_header_format = (  
    ('HEADER ID',1,'<','B',True),
    ('DATA SOURCE ID',1,'<','B',True),
    ('N BYTES IN ENSEMBLE',2,'<','H',True),
    ('SPARE',1,'<','B',True),
    ('N DATA TYPES',1,'<','B',True),
)


fixed_leader_format = (
    ('FIXED LEADER ID',2,'<','H',True),
    ('CPU F/W VER.',1,'<','B',True),
    ('CPU F/W REV.',1,'<','B',True),
    ('SYSTEM CONFIGURATION',2,'<','B',False),
    ('REAL/SIM FLAG',1,'<','B',True),
    ('LAG LENGTH',1,'<','B',True),
    ('NUMBER OF BEAMS',1,'<','B',True),
    ('NUMBER OF CELLS {WN}',1,'<','B',True),
    ('PINGS PER ENSEMBLE {WP}',2,'<','H',True),
    ('DEPTH CELL LENGTH {WS}',2,'<','H',True),
    ('BLANK AFTER TRANSMIT {WF}',2,'<','H',True),
    ('PROFILING MODE {WM}',1,'<','B',True),
    ('LOW CORR THRESH {WC}',1,'<','B',True),
    ('NO. CODE REPS',1,'<','B',True),
    ('%GD MINIMUM {WG}',1,'<','B',True),
    ('ERROR VELOCITY MAXIMUM {WE}',2,'<','H',True),
    ('TPP MINUTES',1,'<','B',True),
    ('TPP SECONDS',1,'<','B',True),
    ('TPP HUNDREDTHS {TP}',1,'<','B',True),
    ('COORDINATE TRANSFORM {EX}',1,'<','B',False),
    ('HEADING ALIGNMENT {EA}',2,'<','H',True),
    ('HEADING BIAS {EB}',2,'<','H',True),
    ('SENSOR SOURCE {EZ}',1,'<','B',True),
    ('SENSORS AVAILABLE',1,'<','B',True),
    ('BIN 1 DISTANCE',2,'<','H',True),
    ('XMIT PULSE LENGTH BASED ON {WT}',2,'<','H',True),
    ('starting cell WP REF LAYER AVERAGE {WL} ending cell',2,'<','B',True),
    ('FALSE TARGET THRESH {WA}',1,'<','B',True),
    ('SPARE1',1,'<','B',False),
    ('TRANSMIT LAG DISTANCE',2,'<','H',True),
    ('CPU BOARD SERIAL NUMBER',8,'<','Q',False),
    ('SYSTEM BANDWIDTH {WB}',2,'<','H',True),
    ('SYSTEM POWER {CQ}',1,'<','B',True),
    ('SPARE2',1,'<','B',False),
    ('INSTRUMENT SERIAL NUMBER',4,'<','I',True),
    ('BEAM ANGLE',1,'<','B',True)
)


variable_leader_format = (
    ('VARIABLE LEADER ID',2,'<','H',True),
    ('ENSEMBLE NUMBER',2,'<','H',True),
    ('RTC YEAR {TS}',1,'<','B',True),
    ('RTC MONTH {TS}',1,'<','B',True),
    ('RTC DAY {TS}',1,'<','B',True),
    ('RTC HOUR {TS}',1,'<','B',True),
    ('RTC MINUTE {TS}',1,'<','B',True),
    ('RTC SECOND {TS}',1,'<','B',True),
    ('RTC HUNDREDTHS {TS}',1,'<','B',True),
    ('ENSEMBLE # MSB',1,'<','B',True),
    ('BIT RESULT',2,'<','H',True),
    ('SPEED OF SOUND {EC}',2,'<','H',True),
    ('DEPTH OF TRANSDUCER {ED}',2,'<','H',True),
    ('HEADING {EH}',2,'<','H',True),
    ('PITCH TILT 1 {EP}',2,'<','h',True),
    ('ROLL TILT 2 {ER}',2,'<','h',True),
    ('SALINITY {ES}',2,'<','H',True),
    ('TEMPERATURE {ET}',2,'<','h',True),
    ('MPT MINUTES',1,'<','B',True),
    ('MPT SECONDS',1,'<','B',True),
    ('MPT HUNDREDTHS',1,'<','B',True),
    ('HDG STD DEV',1,'<','B',True),
    ('PITCH STD DEV',1,'<','B',True),
    ('ROLL STD DEV',1,'<','B',True),
    ('ADC CHANNEL 0',1,'<','B',True),
    ('ADC CHANNEL 1',1,'<','B',True),
    ('ADC CHANNEL 2',1,'<','B',True),
    ('ADC CHANNEL 3',1,'<','B',True),
    ('ADC CHANNEL 4',1,'<','B',True),
    ('ADC CHANNEL 5',1,'<','B',True),
    ('ADC CHANNEL 6',1,'<','B',True),
    ('ADC CHANNEL 7',1,'<','B',True),
    ('ERROR STATUS WORD ESW {CY}',4,'<','I',True),
    ('SPARE1',2,'<','B',False),
    ('PRESSURE',4,'<','I',True),
    ('PRESSURE SENSOR VARIANCE',4,'<','I',True),
    ('SPARE2',1,'<','B',False),
    ('RTC CENTURY',1,'<','B',True),
    ('RTC YEAR',1,'<','B',True),
    ('RTC MONTH',1,'<','B',True),
    ('RTC DAY',1,'<','B',True),
    ('RTC HOUR',1,'<','B',True),
    ('RTC MINUTE',1,'<','B',True),
    ('RTC SECOND',1,'<','B',True),
    ('RTC HUNDREDTH',1,'<','B',True)
)


bottom_track_format = (
    ('BOTTOM-TRACK ID',2,'<','H',True),
    ('BT PINGS PER ENSEMBLE {BP}',2,'<','H',True),
    ('BT DELAY BEFORE RE-ACQUIRE {BD}',2,'<','H',True),
    ('BT CORR MAG MIN {BC}',1,'<','B',True),
    ('BT EVAL AMP MIN {BA}',1,'<','B',True),
    ('BT PERCENT GOOD MIN {BG}',1,'<','B',True),
    ('BT MODE {BM}',1,'<','B',True),
    ('BT ERR VEL MAX {BE}',2,'<','H',True),
    ('Reserved',4,'<','H',True),
    ('BEAM#1 BT RANGE',2,'<','H',True),
    ('BEAM#2 BT RANGE',2,'<','H',True),
    ('BEAM#3 BT RANGE',2,'<','H',True),
    ('BEAM#4 BT RANGE',2,'<','H',True),
    ('BEAM#1 BT VEL',2,'<','H',True),
    ('BEAM#2 BT VEL',2,'<','H',True),
    ('BEAM#3 BT VEL',2,'<','H',True),
    ('BEAM#4 BT VEL',2,'<','H',True),
    ('BEAM#1 BT CORR.',1,'<','H',True),
    ('BEAM#2 BT CORR.',1,'<','H',True),
    ('BEAM#3 BT CORR.',1,'<','H',True),
    ('BEAM#4 BT CORR.',1,'<','H',True),
    ('BEAM#1 EVAL AMP',1,'<','H',True),
    ('BEAM#2 EVAL AMP',1,'<','H',True),
    ('BEAM#3 EVAL AMP',1,'<','H',True),
    ('BEAM#4 EVAL AMP',1,'<','H',True),
    ('BEAM#1 BT %GOOD',1,'<','H',True),
    ('BEAM#2 BT %GOOD',1,'<','H',True),
    ('BEAM#3 BT %GOOD',1,'<','H',True),
    ('BEAM#4 BT %GOOD',1,'<','H',True),
    ('REF LAYER MIN {BL}',2,'<','H',True),
    ('REF LAYER NEAR {BL}',2,'<','H',True),
    ('REF LAYER FAR {BL}',2,'<','H',True),
    ('BEAM#1 REF LAYER VEL',2,'<','H',True),
    ('BEAM #2 REF LAYER VEL',2,'<','H',True),
    ('BEAM #3 REF LAYER VEL',2,'<','H',True),
    ('BEAM #4 REF LAYER VEL',2,'<','H',True),
    ('BM#1 REF CORR',1,'<','H',True),
    ('BM#2 REF CORR',1,'<','H',True),
    ('BM#3 REF CORR',1,'<','H',True),
    ('BM#4 REF CORR',1,'<','H',True),
    ('BM#1 REF INT',1,'<','H',True),
    ('BM#2 REF INT',1,'<','H',True),
    ('BM#3 REF INT',1,'<','H',True),
    ('BM#4 REF INT',1,'<','H',True),
    ('BM#1 REF %GOOD',1,'<','H',True),
    ('BM#2 REF %GOOD',1,'<','H',True),
    ('BM#3 REF %GOOD',1,'<','H',True),
    ('BM#4 REF %GOOD',1,'<','H',True),
    ('BT MAX. DEPTH {BX}',2,'<','H',True),
    ('BM#1 RSSI AMP',1,'<','H',True),
    ('BM#2 RSSI AMP',1,'<','H',True),
    ('BM#3 RSSI AMP',1,'<','H',True),
    ('BM#4 RSSI AMP',1,'<','H',True),
    ('GAIN',1,'<','H',True),
    ('*SEE BYTE 17',1,'<','H',True),
    ('*SEE BYTE 19',1,'<','H',True),
    ('*SEE BYTE 21',1,'<','H',True),
    ('*SEE BYTE 23',1,'<','H',True),
    ('RESERVED',4,'<','H',True),
)        

address_offsets_format = (None,2,'<','H',True)
data_ID_code_format = (None,2,'<','H',True)
echo_intensity_format= (None,1,'<','<H',True)
velocity_format = (None,2,'<','h',True)
coor_mag_format = (None,1,'<','H',True)
pct_good_format = (None,1,'<','H',True)


#%%
class Pd0:
    """
    A class for parsing and managing data from pd0 files, which are commonly used in
    oceanographic data collection with ADCP (Acoustic Doppler Current Profiler) instruments.

    This class provides functionality to read, process, validate, and access ensemble data
    from pd0 files. It includes methods for navigating through the data, retrieving specific
    ensembles, and extracting metadata and hydrodynamic measurements such as velocity and echo intensity.

    Attributes:
        _filepath (str): Path to the pd0 file being processed.
        _checksum (NoneType or bytearray): Checksum used for verifying the integrity of read data.
        _ensemble_rollover (int): Counter to manage the rollover of ensemble indices due to limitations in counting.
        _address_offsets (list of int): List of byte offsets for accessing specific data blocks within the file.
        ensemble_data (list of dicts): Parsed ensemble data stored as dictionaries.
        metadata (metadata): An instance of the nested metadata class containing detailed information about the file.

    Methods:
        __init__(filepath, **kwargs): Constructor for initializing a Pd0 instance.
        __get_first_and_last_ensembles(): Retrieves the first and last ensembles to determine the scope of the data.
        __skip_to_ensemble(n): Skips to a specified ensemble number within the file.
        __find_previous_ensemble(dist, maxiter, offset): Searches backwards to find the start of a valid ensemble.
        __find_next_ensemble(dist, maxiter, offset): Searches forwards to find the start of a valid ensemble.
        __check_valid_ensemble(ensemble): Checks the validity of an ensemble based on predefined criteria.
        __read_ensembles(start, end, print_progress): Reads a range of ensembles from the file.
        __read_ensemble(): Reads and parses a single ensemble from the file.
        __read_ensemble_header(): Reads and parses the header of an ensemble.
        __read_fixed_leader(): Reads and parses the fixed leader of an ensemble.
        __read_variable_leader(): Reads and parses the variable leader of an ensemble.
        __read_next_bytes(fmt): Reads bytes from the file according to a specified format.
        __get_LE_bit_string(byte): Converts a byte into a string of bits (little-endian).
        __parse_system_configuration(syscfg): Parses system configuration data.
        __parse_EX_command(ex): Parses coordinate transformation commands.
        write_pd0(fname): Writes parsed data back to a new pd0 file.
        __write_ensemble(ensemble): Writes a single ensemble's data to a file.
        __write_next_bytes(x, fmt): Writes data to a file in a specified byte format.
    """
    
    
    def __init__(self,filepath,**kwargs):

        
        self._filepath = filepath # filepath for general use across methods
        self._checksum = None # checksum for general use across methods
        self._ensemble_rollover = 0 # counter for ensemble rollover (at 65,536)
        self._address_offsets = [] # for general use across methods
        self.ensemble_data = []
        self.name = kwargs.pop('name',filepath.split(os.sep)[-1])
 
        class metadata:
            def __init__(self,pd0):
                #ef,el = pd0.__get_first_and_last_ensembles()
                self.filepath = filepath
                self.filesize= os.path.getsize(filepath)
                self.first_ensemble_in_file= ef['VARIABLE LEADER']['ENSEMBLE NUMBER']
                self.last_ensemble_in_file = el['VARIABLE LEADER']['ENSEMBLE NUMBER']
                
                
                self.ensemble_size = ef['ENSEMBLE HEADER']['N BYTES IN ENSEMBLE']
                
                self.first_date_in_file = self.__format_ensemble_datetime(ef)
                self.last_date_in_file = self.__format_ensemble_datetime(el)
                
                # ## calculate number of ensembles in the file
                # # case where the ensemble count rolled over in the file
                if self.first_ensemble_in_file > self.last_ensemble_in_file:
                     self.n_ensembles_in_file = (65536 - self.first_ensemble_in_file) + self.last_ensemble_in_file
                else:   
                    self.n_ensembles_in_file = self.last_ensemble_in_file - self.first_ensemble_in_file + 1   
                
            def __format_ensemble_datetime(self,ensemble):
                
                    century = str(ensemble['VARIABLE LEADER']['RTC CENTURY'])
                    if century == '0': century = '20' #Catch for UNKNOWN bug that occures when parsing DVS instruments
                    year = int(century + str(ensemble['VARIABLE LEADER']['RTC YEAR {TS}']))
                    month= ensemble['VARIABLE LEADER']['RTC MONTH {TS}']
                    day = ensemble['VARIABLE LEADER']['RTC DAY {TS}']
                    hour = ensemble['VARIABLE LEADER']['RTC HOUR {TS}']
                    minute = ensemble['VARIABLE LEADER']['RTC MINUTE {TS}']
                    second = ensemble['VARIABLE LEADER']['RTC SECOND {TS}']
                    hundredth = ensemble['VARIABLE LEADER']['RTC HUNDREDTHS {TS}']
                    try:
                        dtime = datetime.datetime(year,month,day,hour,minute,second,hundredth*1000) 
                    except: 
                        dtime = np.nan
                    return dtime
                        
                        
                        
        ef,el = self.__get_first_and_last_ensembles()
        self.metadata = metadata(self)
   
    
        import_file = kwargs.pop('import_file',True)
        print_progress = kwargs.pop('print_progress',True)
        #######################################################################
        
        if import_file:

            ## Read the file 
            start = kwargs.pop('start',1)
            end = kwargs.pop('end',self.metadata.n_ensembles_in_file)
            
            self._file = open(self._filepath,"rb")
            self.__read_ensembles(start,end, print_progress = print_progress)
            self._file.close()
            
            del self._file
            
            if print_progress:
                print(f'\rloaded {self._n_ensembles}/{self.metadata.n_ensembles_in_file} ensembles from {self.metadata.filepath.split(os.sep)[-1]}'.ljust(500), end = '\n')
            

                
        

    def __get_first_and_last_ensembles(self):
        #retrieve first and last ensembles
        self._file= open(self._filepath,"rb")

        self._file.seek(self.__find_next_ensemble())

        first_ensemble = self.__read_ensemble()
        self._file.seek(0,os.SEEK_END)
        self._file.seek(self.__find_previous_ensemble())
        last_ensemble = self.__read_ensemble()
        
        self._file.close()
        del self._file


        return first_ensemble,last_ensemble  
    
    
    def __skip_to_ensemble(self,n):
        
        
        
        # identify the ensemble in the file with number n
        def get_relative_ensemble_number():
            '''
            determine the ensemble number relative to the first ensemble in the 
            file. 1 = first ensemble in file. 
            assumes that the file is at the ensemble header byte position. 

            '''
 
            init_pos = self._file.tell()
            ens = self.__read_ensemble()
            self._file.seek(self._file.tell()+self._address_offsets[1])# move to variable leader start byte 
            en = ens['VARIABLE LEADER']['ENSEMBLE NUMBER'] - self.metadata.first_ensemble_in_file + 1
            self._file.seek(init_pos)
            return en


        init_pos = self._file.tell()
        init_guess = self.metadata.ensemble_size*max((n-1),0)
        
    
        
        self._file.seek(init_guess)

        self._file.seek(self.__find_next_ensemble())
        
        
        ## adjust for first ensemble in the file 
        
        en = get_relative_ensemble_number() 
        diff = en - n

        #search = False
        
        maxiter = self.metadata.n_ensembles_in_file
        _iter = 0
        success = True # successful search
        while diff!=0:
            _iter+=1
            #search = True
            if diff>0: 
                #print(diff)
                self._file.seek(self.__find_previous_ensemble(offset = True))
                #print(f'\rb {diff} en = {en}  n = {n}', end = '\r')
                #time.sleep(.1)
            elif diff<0: 
                #print(f'\rf {diff} en = {en} n = {n}', end = '\n')
                self._file.seek(self.__find_next_ensemble(offset = True))
                #time.sleep(.1)
                
                
            en = get_relative_ensemble_number()
            diff = en - n
            
            
            
            
            if _iter > maxiter:
                success = False # search failed
                break
          
         
        if success:
            self._file.seek(self.__find_previous_ensemble()) # adjust for last ensemble read
        else: # if search failed, return to initial file position 
            self._file.seek(init_pos)
            # if search failed, return to start point 
            
        # elif diff>0: self._file.seek(self.__find_next_ensemble(offset = True))
        # else: self._file.seek(self.__find_previous_ensemble())
            
        
        return success
        
        
        
        


    def __find_previous_ensemble(self,dist = 5000, maxiter = 100, offset = False):
        """
        find the position of the first byte of the previous valid ensemble in the 
        file. If no valid ensembles are found, then the position of the beginning of 
        the search file will be returned. 
        
        The file is always returned to its starting position at the time of the 
        function call. 

        Parameters
        ----------
        dist : int, optional
            number of bytes to read in each iteration. This value must be larger 
            than the size of an ensemble to The default is 2000.
        maxiter : int, optional
            numer of iterations to perform. The default is 100.
        offset : bool, optinal
            whether to ignore the first two bytes (case where its on the next ensemble already)

        Returns
        -------
        pos : int
            position of the starting byte for the next valid enseble in the file. 
            If no valid ensembles are found, then the position of the end of the file will be returned.
        """


        init_pos = self._file.tell() # file position at the beginning of the search 
        # start_pos = max(init_pos - dist,0) # starting position in file for the current iteration
        
        if not offset:start_pos = max(init_pos,0) # starting position in file for the current iteration
        else: start_pos = max(init_pos - 2,0) # starting position in file for the current iteration
        valid = False # indicator variable for successful search 
        _iter = 0 # number of iterations 
        while not valid: # while a valid ensemble has not been found,
            self._file.seek(start_pos) # seek to the start position for the iteration
            headers = [i for i in re.finditer(b'\x7f\x7f', self._file.read(dist), re.S)] # read bytes and search for an instance of the ensemble header ID b'\x7f\x7f'
            #print(headers)
            for header in np.flip(headers): # loop over identtified headers 
                pos = start_pos + header.start() # byte position for the beginning of the header 
                self._file.seek(pos)# seek to byte position for the beginning of the header 
                try: 
                    ensemble = self.__read_ensemble()
                    valid = self.__check_valid_ensemble(ensemble)
                    #print('p',ensemble['VARIABLE LEADER']['ENSEMBLE NUMBER'])
                except: valid = False
                if valid: break # break the loop if the ensemble is valid

            # step backwards
            start_pos = max(start_pos - dist,0)
            self._file.seek(start_pos)
            
            # update iteration count
            _iter+=1 
            
            ## other termination conditions
            term_conds = [_iter>=maxiter,
                          self._file.tell()== 0]
            
            
            
            if any(term_conds):
                pos = 0
                #print('maxout')
                break
        
            
        self._file.seek(init_pos) # return file to initial position 
        return pos 


    def __find_next_ensemble(self,dist = 5000, maxiter = 100,offset = False):
        """
        find the position of the first byte of the next valid ensemble in the 
        file. If no valid ensembles are found, then the position of the end of 
        the search file will be returned. 
        
        The file is always returned to its starting position at the time of the 
        function call. 

        Parameters
        ----------
        dist : int, optional
            number of bytes to read in each iteration. This value must be larger 
            than the size of an ensemble to The default is 3000.
        maxiter : int, optional
            numer of iterations to perform. The default is 100.
        offset : bool, optinal
            whether to ignore the first two bytes (case where its on the next ensemble already)

        Returns
        -------
        pos : int
            position of the starting byte for the next valid enseble in the file. 
            If no valid ensembles are found, then the position of the end of the file will be returned.
        """
        # global ensemble2
        # ensemble2 = self.__read_ensemble()
        init_pos = self._file.tell() # file position at the beginning of the search 
        if not offset:start_pos = max(init_pos,0) # starting position in file for the current iteration
        else: start_pos = max(init_pos + 2,0) # starting position in file for the current iteration
        valid = False # indicator variable for successful search 
        _iter = 0 # number of iterations 
        while not valid: # while a valid ensemble has not been found,
            self._file.seek(start_pos) # seek to the start position for the iteration
            headers = [i for i in re.finditer(b'\x7f\x7f', self._file.read(dist), re.S)] # read bytes and search for an instance of the ensemble header ID b'\x7f\x7f'
            for header in headers: # loop over identtified headers 
                pos = start_pos + header.start() # byte position for the beginning of the header 
                self._file.seek(pos)# seek to byte position for the beginning of the header
                
                # ensemble = self.__read_ensemble()
                # valid = self.__check_valid_ensemble(ensemble)
                #ensemble = self.__read_ensemble()
                try: 
 
                    ensemble = self.__read_ensemble()
                    valid = self.__check_valid_ensemble(ensemble)
                    #print(f'pos {pos}',ensemble['VARIABLE LEADER']['ENSEMBLE NUMBER'])
                    
                except: valid = False
                if valid: break # break the loop if the ensemble is valid

            # step backwards
            start_pos = start_pos + dist -1
            self._file.seek(start_pos)
            _iter+=1 # update iteration count 
            #print(start_pos)
            ## other termination conditions
            term_conds = [_iter>maxiter,
                          self._file.tell()>=os.path.getsize(self._filepath)]
        
            if any(term_conds):
                pos = os.path.getsize(self._filepath)
                #print(term_conds)
                break

        self._file.seek(init_pos) # return file to initial position 
        
        
       
        return pos 

    

    def __check_valid_ensemble(self,ensemble):
        """
        Check that an ensemble is valid 

        Parameters
        ----------
        ensemble : dict
            ensemble dictionary returned by __read_ensemble()

        Returns
        -------
        valid : bool
            True if all checks are passed
        """

        
        
        
        valid = True
        
        ## check ensemble header fields 
        if ensemble['ENSEMBLE HEADER']['N DATA TYPES']>10: valid = False     
        if ensemble['ENSEMBLE HEADER']['N DATA TYPES']<1: valid = False
        if ensemble['ENSEMBLE HEADER']['HEADER ID']!=127: valid = False     
        if ensemble['ENSEMBLE HEADER']['DATA SOURCE ID']!=127: valid = False 
        
        # check a random data field 
        if not any(ensemble['CORRELATION MAGNITUDE']): valid = False




        # try to parse the date
        century = str(ensemble['VARIABLE LEADER']['RTC CENTURY'])
        if not century in ['19','20']: century = '20' 
        year = int(century + str(ensemble['VARIABLE LEADER']['RTC YEAR {TS}']))
        month= ensemble['VARIABLE LEADER']['RTC MONTH {TS}']
        day = ensemble['VARIABLE LEADER']['RTC DAY {TS}']
        hour = ensemble['VARIABLE LEADER']['RTC HOUR {TS}']
        minute = ensemble['VARIABLE LEADER']['RTC MINUTE {TS}']
        second = ensemble['VARIABLE LEADER']['RTC SECOND {TS}']
        hundredth = ensemble['VARIABLE LEADER']['RTC HUNDREDTHS {TS}']
        
        try:
            datetime.datetime(year,month,day,hour,minute,second,hundredth*1000)
        except: 
            valid = False
        
        

            
        
        return valid
    
    
    
    def __read_ensembles(self,start = None,end = None,print_progress = True):
        """
        Read ensembles from the pd0 file. all ensembles between start and end
        will be read. 

        Parameters
        ----------
        start : int, optional
            first ensemble to read. If not specified, the first ensemble in the 
            file will be used. 
        end : int, optional
            last ensemble to read. If not specified, the last ensemble in the 
            file will be used. 
        print_progress : bool, optional
            If true a progress bar will be printed. The default is true.


        """
        

        ## set default parameters
        if start == None:
            start = 1 
        if end == None:
            end = self.metadata.n_ensembles_in_file 
        
        #print(start,end)
        ## check that start/end inputs are valid 
        message = f'Invalid start/end ensemble'
        if (start < 1) or (start >= self.metadata.n_ensembles_in_file ):
            raise ValueError(message)
        if (end <= 1) or (end > self.metadata.n_ensembles_in_file ):
            raise ValueError(message)
        if start>end:
            raise ValueError('start ensemble is greater than end ensemble')




    
        


        # # initalize the progress bar
        # if print_progress: ptools.printProgressBar(0, end-start,taskname = f'{self.metadata.filepath.split(os.sep)[-1]}', prefix = 'parsing .pd0')
        
        
        en = start # most recent ensemble number read 
        self._n_ensembles = 0 # number of ensembles read so far
        
        ## open the file for reading 
        self._file= open(self._filepath,"rb")
        self.__skip_to_ensemble(start) # skip to the start ensemble
        
        
        ## read the next ensemble until the end ensemble has been reached
        while en<=end:    
            valid = False # indicator for whther the read was successful
            #print(en)
            try:
                ensemble = self.__read_ensemble()
                valid = self.__check_valid_ensemble(ensemble) # check if it is a valid ensemble 
            except: 
                None
 


            if valid:# if its a valid ensemble update some other stuff
                # account for ensemble rollover
                self.ensemble_data.append(ensemble) # append to global ensemble data
                self._n_ensembles +=1 # update ensemble number counter 
                if print_progress: ptools.printProgressBar(self._n_ensembles-1, end-start,taskname = f'{self.metadata.filepath.split(os.sep)[-1]}', prefix = 'parsing .pd0')## print progress
                en += 1 #ensemble['VARIABLE LEADER']['ENSEMBLE NUMBER'] # update the ensemble number 
                #print(en)
                
            else: # if ensemble is not valid, skip to the next valid ensemble 
            
                if (self.metadata.filesize -self._file.tell())<self.metadata.ensemble_size:
                    self._file.close()
                    break
                else:
                    ## check if we're close to the end of the file 
                    curr_pos = self._file.tell()
                    success = self.__skip_to_ensemble(self._n_ensembles+start) # skip to last known ensemble 
                    self._file.seek(self.__find_next_ensemble(offset = False)) # move forward one ensemble
                
                
                # print(self._n_ensembles+start+1)
                #success = self.__skip_to_ensemble(self._n_ensembles+start+1)
                
                
                # if not success:
                #     print('skipping')
                #     self._file.seek(self.__find_next_ensemble(offset = True))

                

        self.metadata.first_ensemble_read = self.ensemble_data[0]['VARIABLE LEADER']['ENSEMBLE NUMBER']
        self.metadata.last_ensemble_read  = self.ensemble_data[-1]['VARIABLE LEADER']['ENSEMBLE NUMBER'] 
        
        
        
    def __read_ensemble(self):
        ## check if the file is open
        ensemble_start_byte = self._file.tell()
        
        self._checksum = [] #add new entry to the file checksum
        
        ensemble = { 'ENSEMBLE HEADER':{},
                      'FIXED LEADER': {},
                      'VARIABLE LEADER': {},
                      'VELOCITY': [],
                      'CORRELATION MAGNITUDE':[],
                      'ECHO INTENSITY':[],
                      'PERCENT GOOD':[],
                      'BOTTOM TRACK':[],
                      'SYSTEM CONFIGURATION':{},
                      'COORDINATE SYSTEM': {}} 

        ensemble['ENSEMBLE HEADER'] = self.__read_ensemble_header()
        self._file.seek(ensemble_start_byte+self._address_offsets[0])
        ensemble['FIXED LEADER'] = self.__read_fixed_leader()
        self._file.seek(ensemble_start_byte+self._address_offsets[1])
        ensemble['VARIABLE LEADER'] = self.__read_variable_leader()
        
        

        
    
        self._file.seek(ensemble_start_byte+self._address_offsets[2])
        ensemble['ENSEMBLE HEADER']['VELOCITY ID'] = self.__read_next_bytes(data_ID_code_format)#_NextLittleEndianUnsignedShort(file,2,decode = True)
        velocity_data = []
        for cell in range(ensemble['FIXED LEADER']['NUMBER OF CELLS {WN}']):
            cell_data = []
            for beam in range(4):#ensemble['FIXED LEADER']['NUMBER OF BEAMS']):
                cell_data.append(self.__read_next_bytes(velocity_format))
            velocity_data.append(cell_data) 

        ## Parse data type 4 (correlation Magnitude)
        ensemble['ENSEMBLE HEADER']['ID CODE 4'] =self.__read_next_bytes(data_ID_code_format)# _NextLittleEndianUnsignedShort(file,2, decode = True)
        corr_mag_data = []
        
        for cell in range(ensemble['FIXED LEADER']['NUMBER OF CELLS {WN}']):
            cell_data = []
            for beam in range(4):#ensemble['FIXED LEADER']['NUMBER OF BEAMS']):
                cell_data.append(self.__read_next_bytes(coor_mag_format)) #_NextLittleEndianUnsignedShort(file,1,decode = field[3]))
            corr_mag_data.append(cell_data)

            
        ## Parse data type 5 (echo intensity)
        ensemble['ENSEMBLE HEADER']['ID CODE 5'] =self.__read_next_bytes(data_ID_code_format)# _NextLittleEndianUnsignedShort(file,2,decode = True)
        echo_intensity_data = []
        
        for cell in range(ensemble['FIXED LEADER']['NUMBER OF CELLS {WN}']):
            cell_data = []
            for beam in range(4):#ensemble['FIXED LEADER']['NUMBER OF BEAMS']):
                cell_data.append(self.__read_next_bytes(echo_intensity_format))#_NextLittleEndianUnsignedShort(file,1,decode = field[3]))
            echo_intensity_data.append(cell_data)    
         
        ## Parse data type 6 (percent good)
        ensemble['ENSEMBLE HEADER']['ID CODE 6'] = self.__read_next_bytes(data_ID_code_format)#_NextLittleEndianUnsignedShort(file,2, decode = True)
        pct_good_data = []
        
        for cell in range(ensemble['FIXED LEADER']['NUMBER OF CELLS {WN}']):
            cell_data = []
            for beam in range(4):#ensemble['FIXED LEADER']['NUMBER OF BEAMS']):
                cell_data.append(self.__read_next_bytes(pct_good_format))#_NextLittleEndianUnsignedShort(file,1,decode = field[3]))
            pct_good_data.append(cell_data)      
           
        # read bottom track data, if recorded
        if ensemble['ENSEMBLE HEADER']['N DATA TYPES'] ==7:
            bottom_track = {}
            for field in bottom_track_format:  
                bottom_track[field[0]] = self.__read_next_bytes(field)# _NextLittleEndianUnsignedShort(file,field[1],decode = field[3]) 
        else: bottom_track = None          
        

        ensemble['VELOCITY'] = velocity_data
        ensemble['CORRELATION MAGNITUDE']= corr_mag_data
        ensemble['ECHO INTENSITY']=echo_intensity_data
        ensemble['PERCENT GOOD']=pct_good_data
        ensemble['BOTTOM TRACK']=bottom_track
   
        
        
        #read reserved bit data 
        ensemble['ENSEMBLE HEADER']['RESERVED BIT DATA'] = self.__read_next_bytes(data_ID_code_format)#_NextLittleEndianUnsignedShort(file,2,decode = True)
       
        # read end of ensemble checksum 
        file_checksum = struct.unpack('<H',self._file.read(2))[0] #chacksum to verify

        
        
        passed_checksum = True
        diff = sum(bytearray(b''.join(self._checksum)))%65536 - file_checksum
        if diff != 0:
            #print(f'checksum failed reading ensemble {ensemble["VARIABLE LEADER"]["ENSEMBLE NUMBER"]} \n {diff} byte discrepancy')
            passed_checksum = False
            
           
           
        ## get system configuration parameters 
        try:
            system_configuration = self.__parse_system_configuration(ensemble['FIXED LEADER']['SYSTEM CONFIGURATION'])  
        except: 
            system_configuration = None
           
        coord_sys = self.__parse_EX_command(ensemble['FIXED LEADER']['COORDINATE TRANSFORM {EX}'])
        

        ensemble['SYSTEM CONFIGURATION']=system_configuration
        ensemble['COORDINATE SYSTEM']=coord_sys
        
        return ensemble
        
    def __read_ensemble_header(self):
        #assumes file is already in starting position 
        ensemble_header = {}
        #self.__skip_to_next_header()
        for field in ensemble_header_format:
            ensemble_header[field[0]] = self.__read_next_bytes(field)
           
        ensemble_header['N DATA TYPES'] = max(ensemble_header['N DATA TYPES'],6)    
            
        ensemble_header['ADDRESS OFFSETS'] = []
        for n in range(ensemble_header['N DATA TYPES']):
            ensemble_header['ADDRESS OFFSETS'].append(self.__read_next_bytes(address_offsets_format))
        self._address_offsets = ensemble_header['ADDRESS OFFSETS']

        return ensemble_header
    
    def __read_fixed_leader(self):
        #assumes file is already in starting position 
        fixed_leader = {}
        for field in fixed_leader_format:
            fixed_leader[field[0]] = self.__read_next_bytes(field)
        return fixed_leader
    
    def __read_variable_leader(self):  
        #assumes file is already in starting position
        variable_leader = {}
        for field in variable_leader_format:
                variable_leader[field[0]] = self.__read_next_bytes(field)
                
                
        # ## adjust for ensemble rollover 
        rollover = variable_leader['ENSEMBLE # MSB']
        variable_leader['ENSEMBLE NUMBER'] += rollover*65535


        return variable_leader 
            

                     
    def __read_next_bytes(self,fmt):
        decode = fmt[4]
        raw_bytes = self._file.read(fmt[1])
        # n_bytes_read += n
        self._checksum.append(raw_bytes)
        
        if fmt[1] > 1:
            if fmt[3] =='B':
                fmtstr = f'{fmt[2]}{fmt[1]}{fmt[3]}' 
                
            elif fmt[3] in ['Q','I']:
                fmtstr = f'{fmt[2]}{fmt[3]}' 
                
            else:
                fmtstr = f'{fmt[2]}{int(fmt[1]/2)}{fmt[3]}'
                
            out = struct.unpack(fmtstr,raw_bytes)
        else:
            out = raw_bytes
            
        if decode:    
            return out[0]
        else:
            return raw_bytes
        
        
    def __get_LE_bit_string(self,byte): 
        """
        make a bit string from little endian byte
        
        Args:
            byte: a byte
        Returns:
            a string of ones and zeros, the bits in the byte
        """
        # surely there's a better way to do this!!
        bits = ""
        for i in [7, 6, 5, 4, 3, 2, 1, 0]:  # Little Endian
            if (byte >> i) & 1:
                bits += "1"
            else:
                bits += "0"
        return bits     
    
          
    def __parse_system_configuration(self,syscfg):
        """
        determine the system configuration parameters from 2-byte hex
        
        Args:
            syscfg: 2-byte hex string 
        Returns:
            dictionary of system configuration parameters
        """    
    
        LSB = self.__get_LE_bit_string(syscfg[0])
        MSB = self.__get_LE_bit_string(syscfg[1])
        
        ## determine system configuration
        #key for Beam facing
        beam_facing = {'0':'DOWN',
                    '1':'UP'}
        
        #key for XDCR attached
        xdcr_att = {'0':'NOT ATTACHED',
                    '1':'ATTACHED'}
        
        # key for sensor configuration
        sensor_cfg = {'00':'#1',
                      '01':'#2',
                      '10':'#3'}
        
        # key for beam pattern
        beam_pat = {'0':'CONCAVE',
                    '1':'CONVEX'}
        
        # key for system frequencies
        sys_freq = {'000':'75-kHz',
                    '001':'150-kHz',
                    '010':'300-kHz',
                    '011':'600-kHz',
                    '100':'1200-kHz',
                    '101':'2400-kHz',} 
        
        ## determine system configuration from MSB
        janus = {'0100':'4-BM',
                '0101': '5-BM (DEMOD)',
                '1111': '5-BM (2 DEMD)'}
        
        beam_angle = {'00': '15E',
                     '01': '20E',
                     '10': '30E',
                     '11': 'OTHER'}
        
        system_configuration = {}
        system_configuration['BEAM FACING']  = beam_facing[LSB[0]]
        system_configuration['XDCR HD']      = xdcr_att[LSB[1]]
        system_configuration['SENSOR CONFIG']= sensor_cfg[LSB[2:4]]
        system_configuration['BEAM PATTERN'] = beam_pat[LSB[4]]
        system_configuration['FREQUENCY']    = sys_freq[LSB[5:]]
        try:
            system_configuration['JANUS CONFIG'] = janus[MSB[:4]]
        except: system_configuration['JANUS CONFIG'] = 'UNKNOWN'
        
        system_configuration['BEAM ANGLE']   = beam_angle[MSB[-2:]]
    
        return system_configuration   
    
    
    def __parse_EX_command(self,ex):
        """
        determine the coordinate transformation processing parameters (EX command). parameters from 1-byte hex
        
        Args:
            ex: 1-byte hex string 
        Returns:
            string - coordinate transformation processing parameter 
        """      
    
        LSB = self.__get_LE_bit_string(ex[0])
        
        coord_sys = {'00': 'BEAM COORDINATES',
                     '01': 'INSTRUMENT COORDINATES',
                     '10': 'SHIP COORDINATES',
                     '11': 'EARTH COORDINATES'}    
        
        coord_system= coord_sys[LSB[3:5]]
        
        return coord_system

 
    
    def write_pd0(self,fname):
        self._file= open(f'{fname}.000','wb')  
    # __write_next_bytes(127,ensemble_header_format[0],file)

        for e,ensemble in enumerate(self.ensemble_data):
            self.__write_ensemble(ensemble)
        self._file.close()   




    def __write_ensemble(self,ensemble): 
        self._checksum = bytearray(0)
        
        # # write the ensemble header 
        for field in ensemble_header_format:
            self.__write_next_bytes(ensemble['ENSEMBLE HEADER'][field[0]],field)
            
        for n in range(ensemble['ENSEMBLE HEADER']['N DATA TYPES']):
            self.__write_next_bytes(ensemble['ENSEMBLE HEADER']['ADDRESS OFFSETS'][n],address_offsets_format)
        
            
        #Write the fixed leader (data type 1)
        for field in fixed_leader_format:
            self.__write_next_bytes(ensemble['FIXED LEADER'][field[0]],field)
            
        for field in variable_leader_format:
            self.__write_next_bytes(ensemble['VARIABLE LEADER'][field[0]],field)
               
        ## Write data type 3 (velocity)
        self.__write_next_bytes(ensemble['ENSEMBLE HEADER']['VELOCITY ID'],data_ID_code_format)
        for cell in range(ensemble['FIXED LEADER']['NUMBER OF CELLS {WN}']):
            cell_data = []
            for beam in range(ensemble['FIXED LEADER']['NUMBER OF BEAMS']):
                self.__write_next_bytes(ensemble['VELOCITY'][cell][beam],velocity_format)
        
        ## Write data type 4 (correlation Magnitude)
        self.__write_next_bytes(ensemble['ENSEMBLE HEADER']['ID CODE 4'],data_ID_code_format)
        for cell in range(ensemble['FIXED LEADER']['NUMBER OF CELLS {WN}']):
            cell_data = []
            for beam in range(ensemble['FIXED LEADER']['NUMBER OF BEAMS']):
                self.__write_next_bytes(ensemble['CORRELATION MAGNITUDE'][cell][beam],coor_mag_format)
                
        ## Write data type 5 (echo intensity)
        self.__write_next_bytes(ensemble['ENSEMBLE HEADER']['ID CODE 5'],data_ID_code_format)
        for cell in range(ensemble['FIXED LEADER']['NUMBER OF CELLS {WN}']):
            cell_data = []
            for beam in range(ensemble['FIXED LEADER']['NUMBER OF BEAMS']):
                self.__write_next_bytes(ensemble['ECHO INTENSITY'][cell][beam],echo_intensity_format)
        
        ## Write data type 6 (percent good)
        self.__write_next_bytes(ensemble['ENSEMBLE HEADER']['ID CODE 6'],data_ID_code_format)
        for cell in range(ensemble['FIXED LEADER']['NUMBER OF CELLS {WN}']):
            cell_data = []
            for beam in range(ensemble['FIXED LEADER']['NUMBER OF BEAMS']):
                self.__write_next_bytes(ensemble['PERCENT GOOD'][cell][beam],pct_good_format)     
          
        # write bottom track data, if recorded
        if ensemble['BOTTOM TRACK']:
            if ensemble['ENSEMBLE HEADER']['N DATA TYPES'] ==7:
                for field in bottom_track_format:  
                    self.__write_next_bytes(ensemble['BOTTOM TRACK'][field[0]],field) 
            
        
        #write reserved bit data 
        self.__write_next_bytes(ensemble['ENSEMBLE HEADER']['RESERVED BIT DATA'],data_ID_code_format)
        
        # write end of ensemble checksum 
        self.__write_next_bytes(sum(self._checksum)%65536,data_ID_code_format) #chacksum to verify
        
    
    
    def __write_next_bytes(self,x,fmt):
        decoded = fmt[4]
        if decoded:   
            if fmt[3] == 'h':
                write_bytes = x.to_bytes(fmt[1], byteorder='little',signed = True)
            else:
                write_bytes = x.to_bytes(fmt[1], byteorder='little',signed = False)
        else:
            write_bytes = x
            
        # n_bytes_written += n    
        self._checksum += bytearray(write_bytes)  
        self._file.write(write_bytes)   
        
###############################################################################
#%% Analytics Wrapper 

class DataSet(Pd0):
    """
    Extends the Pd0 class to provide enhanced analytical capabilities specifically
    tailored for handling and analyzing ADCP data.

    This class offers methods for extracting time series data from variable and fixed
    leaders, computing derived quantities like velocities and displacements, and applying
    corrections such as timezone or magnetic deviations. It also facilitates the integration
    with other data types like PT3 data for comprehensive analyses.

    Attributes:
        timezone_correction (int): Hours to adjust the time data for local timezone corrections.
        magnetic_deviation_correction (float): Degrees to adjust the velocity data for magnetic deviations.
        name (str): A derived name from the filepath, generally the filename.
        n_ensembles (int): Number of ensembles loaded into memory.
        config (config_parameters): Configuration parameters derived from the ADCP data.
        plot (plotting): A plotting tool associated with the dataset for visual analysis.
        processing (processing): Processing tools for data manipulation and computation.
        geometry (geometry): Geometrical tools for handling spatial dimensions and transformations.
        mask (masking): Masking tools for data cleaning and preparation.
        PT3 (PT3, optional): PT3 data associated with the dataset, if provided.

    Methods:
        __init__(filepath, **kwargs): Initializes the DataSet instance with the specified ADCP data file and optional parameters.
        get_ensemble_datetimes(): Returns an array of datetime objects representing the timestamps of each ensemble.
        get_leader_data(field, leader='VARIABLE LEADER', dtype=int): Retrieves time series data for a specified field from the leader data.
        get_ensemble_numbers(): Returns an array of ensemble numbers.
        get_sensor_transmit_pulse_lengths(): Retrieves the ADCP transmit pulse lengths.
        get_sensor_temperature(): Retrieves the temperature data from the sensor.
        get_bit_result(): Retrieves the built-in test result from the ADCP.
        get_salinity(): Retrieves the salinity data from the sensor.
        get_sensor_pitch(): Retrieves the pitch data from the sensor.
        get_sensor_roll(): Retrieves the roll data from the sensor.
        get_sensor_heading(): Retrieves the heading data from the sensor.
        get_sensor_transducer_depth(): Retrieves the depth of the transducer below the water surface.
        get_velocity(field_name='VELOCITY'): Retrieves velocity data along with calculated displacements.
        get_bottom_track(): Retrieves bottom tracking data from the ADCP.
        get_ensemble_array(field_name='ECHO INTENSITY', mask=True): Formats specified ensemble data field into a numpy array with optional masking.
    """
    
    def __init__(self,filepath,**kwargs):
        super().__init__(filepath,**kwargs)
        
        
        # parse kwargs
        self.timezone_correction = kwargs.pop('timezone_correction',0)
        self.magnetic_deviation_correction = kwargs.pop('magnetic_deviation_correction',0)  

        self.name = filepath.split(os.sep)[-1]
            
        self.n_ensembles = len(self.ensemble_data)
        #self.set_instrument_config_parameters() # set instrument configuration params
        


        self.config = config_parameters(self)
        self.plot = plotting(self)        
        self.processing = processing(self)
        self.geometry = geometry(self)
        self.mask = masking(self)
        
        
                    
        if kwargs.get('pt3_filepath'):
            self.PT3 = PT3(kwargs.get('pt3_filepath'))

    

    def get_ensemble_datetimes(self,ensemble = None):
        """
        Generate numpy array of datetimes for ensembles. 
        Args:
            None
        Returns:
            numpy array with dimensions (n_ensembles)  
        """           

        dtimes =[]
        for e,ensemble in enumerate(self.ensemble_data):
            century = str(ensemble['VARIABLE LEADER']['RTC CENTURY'])
            if not century in ['19','20']: century = '20' 
            year = int(century + str(ensemble['VARIABLE LEADER']['RTC YEAR {TS}']))
            month= ensemble['VARIABLE LEADER']['RTC MONTH {TS}']
            day = ensemble['VARIABLE LEADER']['RTC DAY {TS}']
            hour = ensemble['VARIABLE LEADER']['RTC HOUR {TS}']
            minute = ensemble['VARIABLE LEADER']['RTC MINUTE {TS}']
            second = ensemble['VARIABLE LEADER']['RTC SECOND {TS}']
            hundredth = ensemble['VARIABLE LEADER']['RTC HUNDREDTHS {TS}']
            try:
                dtimes.append(datetime.datetime(year,month,day,hour,minute,second,hundredth*1000) + datetime.timedelta(hours = self.timezone_correction)) 
            except: 
                print(ensemble['VARIABLE LEADER']['RTC MONTH {TS}'])
                #print(str(ensemble['VARIABLE LEADER']['RTC YEAR {TS}']))
                dtimes.append(datetime.datetime(year,month,day,hour,minute,second,hundredth*1000) + datetime.timedelta(hours = self.timezone_correction)) 
                # dtimes.append(np.nan)
                # break
            
        dtimes = np.array(dtimes)
        return dtimes    


        
    
    def get_leader_data(self,field,leader = 'VARIABLE LEADER', dtype = int):
        
        """
        Get a timeseries of data from fixed leader or variable leader. field 
        names are identical to those in Workhorse Commands and Output Data 
        Format.

        Parameters
        ----------
        field : str
            field name.
        leader : ste, optional
            'FIXED LEADER' or 'VARIABLE LEADER'. The default is 'VARIABLE LEADER'.
        dtype : numpy dtype, optional
            data type for the output numpy array. The default is int.

        Returns
        -------
        data : numpy array
             One value for each ensemble.

        """
        
        data = np.empty(self.n_ensembles, dtype = int)
        for c,ensemble in enumerate(self.ensemble_data):
            data[c] = ensemble[leader][field]
        
        return data 
        
    def get_ensemble_numbers(self):
        """
        Get the ensemble number for every ensemble. 

        Returns
        -------
        ensemble_numbers : numpy array
            ensemble numbers. One value for each ensemble.
            
        """
        ensemble_numbers = self.get_leader_data(field = 'ENSEMBLE NUMBER',leader = 'VARIABLE LEADER')
        return ensemble_numbers

        
    def get_sensor_transmit_pulse_lengths(self):
        """
        Get ADCP transmit pulse lengths. Data is from the ensemble variable headers.

        Returns
        -------
        transmit_pulse_lengths : numpy array
            ADCP transmit pule lengths. One value for each ensemble.

        """
        
        transmit_pulse_lengths = self.get_leader_data(field = 'XMIT PULSE LENGTH BASED ON {WT}',leader = 'FIXED LEADER')/100
        return transmit_pulse_lengths  
    
    def get_sensor_temperature(self):
        """
        Temperature of the water at the transducer head
        (ET-command). This value may be a manual setting or a
        reading from a temperature sensor.
        Scaling: LSD = 0.01 degree; Range = -5.00 to +40.00 degrees
        
        Returns
        -------
        temperature : numpy array
            ADCP temperature data in degrees C. One value for each ensemble.

        """
        temperature = np.empty(self.n_ensembles)
        
        for c,ensemble in enumerate(self.ensemble_data):
            temperature[c] = ensemble['VARIABLE LEADER']['TEMPERATURE {ET}']/100
        return temperature
    
    def get_bit_result(self):
        """
        This field contains the results of the WorkHorse ADCP’s Builtin
        Test function. A zero code indicates a successful BIT
        result.
        
        Returns
        -------
        bit_result : numpy array
            ADCP bit result data. One value for each ensemble.

        """
        bit_result = np.empty(self.n_ensembles)
        
        for c,ensemble in enumerate(self.ensemble_data):
            bit_result[c] = ensemble['VARIABLE LEADER']['BIT RESULT']/100
        return bit_result
    
    def get_salinity(self):
        """
        Salinity value of the water at the transducer
        head (ES-command). This value may be a manual setting or
        a reading from a conductivity sensor.
        Scaling: LSD = 1 part per thousand; Range = 0 to 40 ppt
        
        Returns
        -------
        salinity : numpy array
            One value for each ensemble.

        """
        salinity = np.empty(self.n_ensembles)
        
        for c,ensemble in enumerate(self.ensemble_data):
            salinity[c] = ensemble['VARIABLE LEADER']['SALINITY {ES}']
        return salinity
    
    def get_sensor_pitch(self):
        """
        WorkHorse ADCP pitch angle (EP-command).
        This value may be a manual setting or a reading from a tilt
        sensor. Positive values mean that Beam #3 is spatially
        higher than Beam #4.
        Scaling: LSD = 0.01 degree; Range = -20.00 to +20.00 degrees
        Returns
        -------
        pitch : numpy array
            One value for each ensemble.

        """
        pitch = np.empty(self.n_ensembles)
        for c,ensemble in enumerate(self.ensemble_data):
            pitch[c] = ensemble['VARIABLE LEADER']['PITCH TILT 1 {EP}']/100
        return pitch
    
    def get_sensor_roll(self):
        """
        WorkHorse ADCP roll angle (ER-command).
        This value may be a manual setting or a reading from a tilt
        sensor. For up-facing WorkHorse ADCPs, positive values
        mean that Beam #2 is spatially higher than Beam #1. For
        down-facing WorkHorse ADCPs, positive values mean that
        Beam #1 is spatially higher than Beam #2.
        Scaling: LSD = 0.01 degree; Range = -20.00 to +20.00 degrees

        Returns
        -------
        roll : numpy array
            One value for each ensemble.

        """
        roll = np.empty(self.n_ensembles)
        for c,ensemble in enumerate(self.ensemble_data):
            roll[c] = ensemble['VARIABLE LEADER']['ROLL TILT 2 {ER}']/100
        return roll
    
    def get_sensor_heading(self):
        """
        WorkHorse ADCP heading angle (EHcommand).
        This value may be a manual setting or a reading
        from a heading sensor.
        Scaling: LSD = 0.01 degree; Range = 000.00 to 359.99 degrees
        
        Returns
        -------
        heading : numpy array
             One value for each ensemble.

        """

        heading = np.empty(self.n_ensembles)
        for c,ensemble in enumerate(self.ensemble_data):
            heading[c] = ensemble['VARIABLE LEADER']['HEADING {EH}']/100
        return heading
    
    def get_sensor_transducer_depth(self):
        """
        Depth of the transducer below the water surface
        (ED-command). This value may be a manual setting or a
        reading from a depth sensor.
        Scaling: LSD = 1 decimeter; Range = 1 to 9999 decimeters
        
        Returns
        -------
        transducer_depth : numpy array
             One value for each ensemble.

        """

        transducer_depth = np.empty(self.n_ensembles)
        for c,ensemble in enumerate(self.ensemble_data):
            transducer_depth[c] = ensemble['VARIABLE LEADER']['DEPTH OF TRANSDUCER {ED}']
        

        return transducer_depth
    
    def get_velocity(self,field_name = 'VELOCITY'):

        """
        Retrieve velocity vectors including progressive vectors in m/s.
        Magnetic deviation correction is applied here. 
        Parameters
        -------
        field_name: str, optional
            ensemble field name to use for velocity data. 'VELOCITY' is default. 
            Other options might be 'VELOCITY CORRECTED' or other user specified velocity-like fields
        Returns
        -------
        u : numpy array
            u velocity array with dimensions (n_bins, n_ensembles)
        v : numpy array
            v velocity array with dimensions (n_bins, n_ensembles)
        z : numpy array
            z velocity array with dimensions (n_bins, n_ensembles)
        du : numpy array
            displacement in u direction. Array with dimensions (n_bins, n_ensembles)
        dv : numpy array
            displacement in v direction. Array with dimensions (n_bins, n_ensembles)
        dz : numpy array
            displacement in z direction. Array with dimensions (n_bins, n_ensembles)
        err: numpy array
            error velocity array with dimensions (n_bins, n_ensembles)
    
        """
        

        
        
        u = np.empty((self.n_ensembles,self.config.number_of_cells))
        v = np.empty((self.n_ensembles,self.config.number_of_cells))
        z = np.empty((self.n_ensembles,self.config.number_of_cells))
        ev = np.empty((self.n_ensembles,self.config.number_of_cells))
        for e,ensemble in enumerate(self.ensemble_data):
            # cast velocity data as float
            vel_data = np.array(ensemble[field_name]).astype('float')
            # mark nan values 
            vel_data[vel_data==-32768] = np.nan
            vel_data[vel_data==32768] = np.nan
            
            # convert to m/s
            vel_data = vel_data*.001


            if len(vel_data) == self.config.number_of_cells:
                u[e,:] = vel_data[:,0]
                v[e,:] = vel_data[:,1]
                z[e,:] = vel_data[:,2]
                ev[e,:] = vel_data[:,3]
            else:
                u[e,:] = np.nan
                v[e,:] = np.nan
                z[e,:] = np.nan
                ev[e,:] = np.nan
            

        if self.magnetic_deviation_correction!=0:
            for b in range(self.config.number_of_cells): # loop over all bins and calculate rotated velocites for each
                X = np.array([u[:,b],v[:,b]])
                
                X_rot = ptools.gen_rot_z(self.magnetic_deviation_correction)[:2,:2]
                #print(X_rot.shape,X.shape)
                X_rot = np.dot(X_rot,X)
                # update u and v
                u[:,b] = X_rot[0,:]
                v[:,b] = X_rot[1,:]    
                
        ## gen prog vector components 
        t = self.get_ensemble_datetimes()
        to_sec = np.vectorize(lambda x: x.seconds) # vectorized function to convert datetime delta to total seconds
        dt = np.empty(self.n_ensembles)
        dt[:-1] = to_sec(np.diff(t))
        dt[-1] = dt[-2] #assign last value
        dt = np.outer(dt,np.ones(self.config.number_of_cells))
        du = u*dt
        dv = v*dt
        dz = z*dt
        
        #transpose to correct dimensions and apply active masks 
        # u = self.mask.apply_masks(u.T)
        # v = self.mask.apply_masks(v.T)
        # z = self.mask.apply_masks(z.T)
        # du = self.mask.apply_masks(du.T)
        # dv = self.mask.apply_masks(dv.T)
        # dz = self.mask.apply_masks(dz.T)
        # ev = self.mask.apply_masks(ev.T)
        
        return u,v,z,du,dv,dz,ev     

    def get_bottom_track(self):
        """
        

        Parameters
        ----------
        beam_number : int
            beam number to use. if zero the beam average will be returned.

        Returns
        -------
        bt_range : numpy array
            numpy array of distance to bottom 

        """
        #
        bt_range = np.empty((4,self.n_ensembles))
        for e in range(self.n_ensembles):
            if self.ensemble_data[e]['BOTTOM TRACK']:
                for b in range(self.config.number_of_beams):
                    bt_range[b,e] = self.ensemble_data[e]['BOTTOM TRACK'][f'BEAM#{b+1} BT RANGE']/100
            else: bt_range[:,e] = np.nan
        return bt_range 
    
        #else: print('No Bottom Track Data')
        
        #raise NotImplementedError()


    def get_ensemble_array(self,field_name = 'ECHO INTENSITY', mask = True):
        f"""
        Format ensemble data field into a numpy array. All active masks are applied 
        to the ensemble data. 
    
        Parameters
        ----------
        field_name : str, optional
            Ensemble data type (excluding velocity). ['ECHO INTENSITY',
            'PERCENT GOOD','CORRELATION MAGNITUDE', 'SIGNAL TO NOISE RATIO',
            'ABSOLUTE BACKSCATTER'). The default is 'ECHO INTENSITY'.
        mask : bool, optional 
            If True, all activated masks will be applied to the ensemble array.
            if False, the raw data will be returned. 
    
        Returns
        -------
        X : numpy array
            Array of ensemble data with dimensions (n_beams,n_bins,n_ensembles).
    
        """
        ## Case where all beams are returned 
        X = np.empty((4,self.config.number_of_cells,self.n_ensembles))
        for c,ensemble in enumerate(self.ensemble_data):
            column_data = [i for i in ensemble[field_name]]
            X[:,:,c] = np.transpose(ensemble[field_name])
        
        
        X[abs(X)==32768] = np.nan
        # # apply any masks    
        if mask:
            X = self.mask.apply_masks(X)
        
        return X               
         



############### Misc Functions ################################################
       
 

############### Misc Functions ################################################


############### Masking Functions #############################################        

#%%
class config_parameters:
    def __init__(self,pd0,en = 0):
        self.n_ensembles = len(pd0.ensemble_data)
        self.number_of_beams = pd0.ensemble_data[en]['FIXED LEADER']['NUMBER OF BEAMS']
        self.beam_angle = pd0.ensemble_data[en]['FIXED LEADER']['BEAM ANGLE']
        self.bin_1_distance =pd0.ensemble_data[en]['FIXED LEADER']['BIN 1 DISTANCE']
        
        self.frequency      = pd0.ensemble_data[en]['SYSTEM CONFIGURATION']['FREQUENCY']
        self.beam_facing    = pd0.ensemble_data[en]['SYSTEM CONFIGURATION']['BEAM FACING']
        self.janus_config   = pd0.ensemble_data[en]['SYSTEM CONFIGURATION']['JANUS CONFIG']
        
        self.cpu_board_serial_number=pd0.ensemble_data[en]['FIXED LEADER']['CPU BOARD SERIAL NUMBER']
        self.cpu_fw_rev =pd0.ensemble_data[en]['FIXED LEADER']['CPU F/W REV.']
        self.cpu_fw_ver =pd0.ensemble_data[en]['FIXED LEADER']['CPU F/W VER.']
        self.instrument_serial_number =pd0.ensemble_data[en]['FIXED LEADER']['INSTRUMENT SERIAL NUMBER']
        self.lag_length =pd0.ensemble_data[en]['FIXED LEADER']['LAG LENGTH']
        self.no_code_reps =pd0.ensemble_data[en]['FIXED LEADER']['NO. CODE REPS']
        self.real_sim_flag =pd0.ensemble_data[en]['FIXED LEADER']['REAL/SIM FLAG']
        self.sensors_available =pd0.ensemble_data[en]['FIXED LEADER']['SENSORS AVAILABLE']
        self.system_configuration = pd0.ensemble_data[en]['FIXED LEADER']['SYSTEM CONFIGURATION']
        self.tpp_minutes =pd0.ensemble_data[en]['FIXED LEADER']['TPP MINUTES']
        self.tpp_seconds =pd0.ensemble_data[en]['FIXED LEADER']['TPP SECONDS']
        self.transmit_lag_distance =pd0.ensemble_data[en]['FIXED LEADER']['TRANSMIT LAG DISTANCE']
        
        self.pct_gd_minimum =pd0.ensemble_data[en]['FIXED LEADER']['%GD MINIMUM {WG}']
        self.blank_after_transmit =pd0.ensemble_data[en]['FIXED LEADER']['BLANK AFTER TRANSMIT {WF}']
        self.coordinate_transform =pd0.ensemble_data[en]['FIXED LEADER']['COORDINATE TRANSFORM {EX}']
        self.depth_cell_length =pd0.ensemble_data[en]['FIXED LEADER']['DEPTH CELL LENGTH {WS}'] #/100
        self.error_velocity_maximum =pd0.ensemble_data[en]['FIXED LEADER']['ERROR VELOCITY MAXIMUM {WE}']
        self.false_target_thresh =pd0.ensemble_data[en]['FIXED LEADER']['FALSE TARGET THRESH {WA}']
        self.heading_alignment =pd0.ensemble_data[en]['FIXED LEADER']['HEADING ALIGNMENT {EA}']
        self.heading_bias =pd0.ensemble_data[en]['FIXED LEADER']['HEADING BIAS {EB}']
        self.low_corr_thresh =pd0.ensemble_data[en]['FIXED LEADER']['LOW CORR THRESH {WC}']
        self.number_of_cells =pd0.ensemble_data[en]['FIXED LEADER']['NUMBER OF CELLS {WN}']
        self.pings_per_ensemble =pd0.ensemble_data[en]['FIXED LEADER']['PINGS PER ENSEMBLE {WP}']
        self.profiling_mode =pd0.ensemble_data[en]['FIXED LEADER']['PROFILING MODE {WM}']
        self.sensor_source =pd0.ensemble_data[en]['FIXED LEADER']['SENSOR SOURCE {EZ}']
        self.starting_cell_wp_ref =pd0.ensemble_data[en]['FIXED LEADER']['starting cell WP REF LAYER AVERAGE {WL} ending cell']
        self.system_bandwidth = pd0.ensemble_data[en]['FIXED LEADER']['SYSTEM BANDWIDTH {WB}']
        self.system_power =pd0.ensemble_data[en]['FIXED LEADER']['SYSTEM POWER {CQ}']
        self.tpp_hundredths =pd0.ensemble_data[en]['FIXED LEADER']['TPP HUNDREDTHS {TP}']          
#%%         

# fpath = r'P:\\41806287\\41806287 NORI-D Data\\Data\\Fixed Stations\\02_Fixed_Bottom_Current_Turbidity\\01_FBCT1\\01_ADCP_600kHz-24149\\Raw\\FBCT1_Modem_Download_02112022\\FBCT1_0308_02112022.000'

# fpath = r'P:\\41806287\\41806287 NORI-D Data\\Data\\Fixed Stations\\02_Fixed_Bottom_Current_Turbidity\\01_FBCT1\\01_ADCP_600kHz-24149\\Raw\\FBCT1_Modem_Download_17102022\\FBCT1_0422_-1102022.000'
# # fpath = r'P:\41806287\41806287 NORI-D Data\Data\Fixed Stations\02_Fixed_Bottom_Current_Turbidity\01_FBCT1\01_ADCP_600kHz-24149\Raw\FBCT1_Modem_Download_10102022\FBCT1_0741_10102022.000'
# fpath = r'P:\\41806287\\41806287 NORI-D Data\\Data\\Fixed Stations\\02_Fixed_Bottom_Current_Turbidity\\01_FBCT1\\01_ADCP_600kHz-24149\\Raw\\FBCT1_Modem_Download_19102022\\FBCT1_1911_19102022.000'

# #fpath = r'\\USDEN1-STOR.DHI.DK\Projects\41806287\41806287 NORI-D Data\Data\Fixed Stations\02_Fixed_Bottom_Current_Turbidity\02_FBCT2\01_ADCP_600kHz-24144\Raw\Data Dump 09192022\09192020_2130.000'
# #pd0 = Pd0(fpath, import_file = True)#, start = 2390, end = 16547)

# adcp = DataSet(fpath)

# t = adcp.get_ensemble_datetimes()
#%%
# # # # #fpath = r'\\USDEN1-STOR.DHI.DK\\Projects\\41806287\\Backups\\Raw_Data_Backup_Hidden_Gem\\ROV_data_HG\\ROVDive20220928\\DHIDive08004r.000'
# pd0_fpath = r'\\USDEN1-STOR.DHI.DK\\Projects\\41806287\\41806287 NORI-D Data\\Data\\ROV\\Island Pride\\ADCP\\Raw\\ADCP_24142_600kHz\\ROV_ADCP_27102022\\_RDI_027.000'

# pd0_fpath = r'\\USDEN1-STOR.DHI.DK\\Projects\\41806287\\41806287 NORI-D Data\\Data\\Fixed Stations\\03_Movable_Turbidity\\02_MT2\\01_ADCP_600kHz-24146\\Raw\\MTS2_000.000'
# # pd0 = Pd0(fpath, import_file = False,)# start = 2390, end = 16547)  # 2390
# # # pd0 = Pd0(fpath, import_file = True, start = 2390, end = 16547)  # 2390


# pd0_fpath = 'RDI_Combine_20181127_20190111_export.000'



# # # # pt3_fpath = '\\\\USDEN1-STOR.DHI.DK\\Projects\\41806287\\41806287 NORI-D Data\\Data\\ROV\\Island Pride HD14\\ADCP\\Config\\ROV_ADCP_20161_PT3.txt'


# pd0 = Pd0(pd0_fpath, import_file = True)#, start = 1, end = 5)
# e = pd0.ensemble_data
#%%
# f = open(pd0_fpath,'rb')

# print(f.read(2))
# print(f.read(2))
# print(f.read(2))
# print(f.read(2))
# f.close()

# f = open(pd0_fpath,'rb')


# f.close()
#%%

# #%%
# # pd0_filepath = r'\\USDEN1-STOR.DHI.DK\Projects\41806287\41806287 NORI-D Data\Data\Fixed Stations\02_Fixed_Bottom_Current_Turbidity\01_FBCT1\01_ADCP_600kHz-24149\Raw\FBCT1_Full_Download_17112022\FBCT1000.000'
# # pt3_filepath  =  r'\\USDEN1-STOR.DHI.DK\Projects\41806287\41806287 NORI-D Data\Data\Fixed Stations\02_Fixed_Bottom_Current_Turbidity\01_FBCT1\01_ADCP_600kHz-24149\Config_File\FBCT01_P3_TEST_17112022.txt'


# pd0_filepath = r'\\USDEN1-STOR.DHI.DK\Projects\41806287\41806287 NORI-D Data\Data\Fixed Stations\02_Fixed_Bottom_Current_Turbidity\01_FBCT1\01_ADCP_600kHz-24149\Raw\FBCT1_Full_Download_17112022\FBCT1000.000'
# # # # # pt3_filepath  =  r'\\USDEN1-STOR.DHI.DK\Projects\41806287\41806287 NORI-D Data\Data\Fixed Stations\02_Fixed_Bottom_Current_Turbidity\01_FBCT1\01_ADCP_600kHz-24149\Config_File\FBCT01_P3_TEST_17112022.txt'


# # # pt3_filepath = r'\\USDEN1-STOR.DHI.DK\Projects\41806287\41806287 NORI-D Data\Data\ROV\Island Pride HD14\ADCP\Config\ROV_ADCP_20161_PT3.txt'

# # # # pd0_filepath = r'\\USDEN1-STOR.DHI.DK\Projects\41806287\41806287 NORI-D Data\Data\Fixed Stations\04_Fixed_Turbidity\02_FT2\01_ADCP_600kHz-24157\Raw\FTS2_000.000'
# # # # pt3_filepath = r'\\USDEN1-STOR.DHI.DK\Projects\41806287\41806287 NORI-D Data\Data\Fixed Stations\04_Fixed_Turbidity\02_FT2\01_ADCP_600kHz-24157\Config_File\PT3.txt'
# # # # # pd0_filepath = r'C:\Users\anba\OneDrive - DHI\Desktop\Projects\NORI Post-Campaign\PCV ADCP\ADCP_PCV_Setup\CL16_\CL16_001.000'
# pd0_filepath = r'\\USDEN1-STOR.DHI.DK\Projects\41806287\41806287 NORI-D Data\Data\ROV\Island Pride HD14\ADCP\Raw\ADCP_24142_600kHz\ROV_ADCP_07102022\_RDI_000.000'
# # # # pd0 = Pd0(pd0_fpaths[0],import_file = False)#.write_pd0('test2')  
# # pd0_filepath = r'\\USDEN1-STOR.DHI.DK\\Projects\\41806287\\Backups\\Raw_Data_Backup_Hidden_Gem\\ROV_data_HG\ROVDive20221103\DHIDive28002r.000'                                                                                          
# pd0_filepath = r'\\USDEN1-STOR.DHI.DK\\Projects\\41806287\\Backups\\Raw_Data_Backup_Hidden_Gem\\ROV_data_HG\\ROVDive20221026\\DHIDive26002r.000'


# pd0 = Pd0(pd0_filepath, import_file = False)#,print_progress = False, start = 3535, end= 3539)
# ds = DataSet(pd0_filepath, import_file = True,print_progress = True)#, start = 2189, end= 2189)
# # print(ds.get_ensemble_numbers())
#plt.plot(np.diff(pd0.get_ensemble_numbers()))
# pd0_filepath = r'\\USDEN1-STOR.DHI.DK\\Projects\\41806287\\Backups\\Raw_Data_Backup_Hidden_Gem\\ROV_data_HG\\ROVDive20221021\\DHIDive22005r.000'
# #pd0_filepath = r'\\USDEN1-STOR.DHI.DK\Projects\41806287\41806287 NORI-D Data\Data\ROV\Island Pride HD14\ADCP\Raw\ADCP_24142_600kHz\ROV_ADCP_07102022\_RDI_013.000'
# # # pd0_filepath = r'\\USDEN1-STOR.DHI.DK\\Projects\\41806287\\Backups\\Raw_Data_Backup_Hidden_Gem\\ROV_data_HG\\ROVDive20220810\\test005r.000'
# pd0 = Pd0(pd0_filepath,import_file = False)
# #%%



# -*- coding: utf-8 -*-
"""
Created on Mon May 15 15:35:53 2023

@author: sndn


to do 

method to efficiently sceen the file by only parsing header info
add position data and other metadata to csv output
"""
###############################################################################

import os
import warnings
import numpy as np
import pandas as pd
import xml.etree.ElementTree as ET


import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from matplotlib.collections import LineCollection
import matplotlib.ticker as mticker


import copy


from ..ptools import ptools
from ..plotting.matplotlib_shell import subplots, dhi_colors


np.seterr(divide='ignore', invalid='ignore')
###############################################################################

class seabird_ctd:
    def __init__(self, filepath_hex, filepath_xml,**kwargs):
        
        verbose = kwargs.pop('verbose',0)
        
        self._filepath_hex = filepath_hex
        self._filepath_xml = filepath_xml
        self.name = kwargs.pop('name',filepath_hex.split(os.sep)[-1])
        self.data = {}
        kwargs.pop('name',filepath_hex.split(os.sep)[-1])
        import_file = kwargs.pop('import_file',True)
        if not import_file:
            self.__detect_sensors()
            self.__parse_xml()
            self.__parse_hex()
        else:
            self.__detect_sensors()
            self.__parse_xml()
            self.__parse_hex()
            
            
            if verbose >0 : self.__print_metadata()   # Comment out to mute metadata
            
            self.__get_xml_coefficients()
            
            try:
                self.__get_hex_coefficients()
                self.__compare_coefficients()
            except:
                None
                #print('No hex coefficients found')
            
            
            self.__hex2temperature()
            self.__hex2pressure()
            self.__hex2conductivity()
            self.__conductivity2salinity()
            
            if 'TurbidityMeter' in self._sensors:
                self.__hex2turbidity()
                
            if 'WET_LabsCStar' in self._sensors:
                self.__hex2transmission()
                
            if 'OxygenSensor' in self._sensors:
                self.__hex2oxygen()


    def __detect_sensors(self):
        
        """
        
        Detects all instruments found in the xml configuration file (.xmlcon) and stores their names in self._sensors
        
        """
        
        self._sensors = []
        self._sensor_index = {}
        
        
        
        tree = ET.parse(self._filepath_xml)
        root = tree.getroot()

        sensors = root.findall('Instrument/SensorArray/Sensor')
        
        #get number of external voltage channels 
        if root.find('Instrument/ExternalVoltageChannels'):
            n_ext_sensors = int(root.find('Instrument/ExternalVoltageChannels').text)
        else:
            n_ext_sensors = 0
        
        sensor_idx = 0
        for sensor in sensors:
            for child in sensor:
                if child.tag != 'NotInUse':
                    
                    self._sensors.append(child.tag)
                    
                    
                    # set sensor index in order read from file
                    self._sensor_index[child.tag] = sensor_idx
                    sensor_idx +=1
                    
    def __print_metadata(self):
        
        """
        
        Prints file metadata to standard output (terminal)
        
        """
        
        print('\n#################### Seabird CTD Hex File ####################\n')
        print("\033[4m" + 'METADATA' + "\033[0m" + '\n')
        print('    ' + "\033[4m" + 'Instrument name' + "\033[0m" + f': {self._xmlcon["Name"]}')
        print('    ' + "\033[4m" + 'File name' + "\033[0m" + f': {os.path.basename(self._filepath_hex)}')
        print('    ' + "\033[4m" + 'File size' + "\033[0m" + f': {os.path.getsize(self._filepath_hex):,} bytes')
        print('    ' + "\033[4m" + 'Number samples' + "\033[0m" + f': {self.n_samples:,}')
        
        print('    ' + "\033[4m" + 'Attached sensors'+ "\033[0m" + ':')
        for sensor in self._sensors:
            print(f'        {sensor}')
        

    def __parse_xml(self):
        
        """
        
        Extracts information from the xml configuration file (.xmlcon) and stores it in self._xmlcon dictionary
        
        """
        global root,tree
        self._xmlcon = {}
        
        tree = ET.parse(self._filepath_xml)
        root = tree.getroot()
        
        self._xmlcon['Name'] = root.find('Instrument/Name').text
        
        self._xmlcon['Sampling Interval'] = float(root.find('Instrument/SampleIntervalSeconds').text)
    

    def __parse_hex(self):
        
        """
        
        Extracts information from SBE16/19 hex files such as hexadecimal strings (stored in self._hex_samples)
        and number of samples.
        
        Constructs timestamps for each sample and stores them in self.data['DateTime']
        
        """
        
        
        hex_file = open(self._filepath_hex, 'r')
        hex_data = hex_file.read()
        hex_file.close()
        
        self._hex_samples = hex_data.split('*END*')[1].strip().split('\n')
        a = self._hex_samples
        
        hex_header = hex_data.split('*END*')[0].strip().split('\n')
        

        time_start = []         # Array to store start times for each sampling period in the file 
        number_samples = []     # Array to store number of samples for each sampling period in the file
  
        
        found= False # indicator if the * hdr was found (for method 1)

        ## method 1 (works for IP instruments)
        for line in hex_header:
            if '* hdr' in line or '* cast' in line:
                found = True
                time_start.append(pd.to_datetime(line.split('samples')[0][-21:]))
                sample_range = [int(i) for i in line.split('samples')[1].split(',')[0].split('to')]
                n_samples = (sample_range[1] - sample_range[0]) + 1
                number_samples.append(n_samples)
                
                sampling_freq = 1/(int(.1*self._xmlcon['Sampling Interval']))
                #print(sampling_freq)
                timestep = pd.Timedelta(value=1/sampling_freq, unit='seconds')
                
                
                if 'int' in line:
                    timestep = int(line.split('int = ')[1].split(',')[0])
                    timestep = pd.Timedelta(timestep, unit = 'seconds')
                    
                    
                date_time = pd.date_range(start=time_start[0], periods=0, freq=timestep) # Create empty DatetimeIndex 
                for i in range(len(time_start)):
                    daterange = pd.date_range(start=time_start[i], periods=number_samples[i], freq=timestep)
                    date_time = date_time.append(daterange)
                
                self.n_samples = sum(number_samples)
                self.data['DateTime'] = date_time
                    
  
        if not found:
            #method 2 (works for HG instruments)
            #Procedure for parsing DeckSamplingSBE hex headers

            n_samples = len(self._hex_samples) 

            hex_header = hex_data.split('*END*')[0]
            time_start = pd.to_datetime(hex_header.split('* System UTC = ')[1])
            time_step = float(hex_header.split('* Real-Time Sample Interval = ')[1].split('seconds')[0])
            time_step = pd.Timedelta(value=time_step, unit='seconds')
            self.n_samples = n_samples
            self.data['DateTime'] = pd.date_range(start=time_start, periods=n_samples, freq=time_step)
        
          
        
        
    def __get_xml_coefficients(self):
        
        """
        
        Retrieves calibration coefficients for sensors detected in xml configuration file (.xmlcon) and
        stores them in self._coefficients_xml.
        
        """
        
        # Dictionary of all possible sensors (as they appear in xml) and their calibration coefficients (the ones you want).
        sensor_coefficients = \
        {
        'TemperatureSensor': ['A0','A1','A2','A3'],
        'ConductivitySensor': ['G','H','I','J','CPcor','CTcor'],
        'PressureSensor': ['PA0','PA1','PA2','PTEMPA0','PTEMPA1','PTEMPA2','PTCA0','PTCA1','PTCA2','PTCB0','PTCB1','PTCB2'],
        'WET_LabsCStar': ['M','B', 'PathLength'],
        'OxygenSensor': ['A','B','C','D0','D1','D2','E','Tau20','H1','H2','H3','Soc','offset'],
        'TurbidityMeter': ['ScaleFactor','DarkVoltage']
        }
        
        tree = ET.parse(self._filepath_xml)
        root = tree.getroot()
        
        self._coefficients_xml = {}
        
        for sensor in self._sensors:
            
            self._coefficients_xml[sensor] = {}
            
            if sensor == 'ConductivitySensor':
                child = root.findall('Instrument/SensorArray/Sensor/ConductivitySensor/Coefficients')[1]
            
            elif sensor == 'OxygenSensor' and self._xmlcon['Name'] == 'SBE 19plus V2 Seacat CTD':
                child = root.findall('Instrument/SensorArray/Sensor/OxygenSensor/CalibrationCoefficients')[1]
                
            elif sensor == 'OxygenSensor' and self._xmlcon['Name'] == 'SBE 16plus V2 Seacat CTD':
                # Overwrite OxygenSensor coefficients if SBE16 (uses different DO sensor and calibration)
                sensor_coefficients['OxygenSensor'] = ['A0','A1','A2','B0','B1','C0','C1','C2', 'pcor']
                child = root.find('Instrument/SensorArray/Sensor/OxygenSensor')
            
            else:
                child = root.find(f'Instrument/SensorArray/Sensor/{sensor}')
            
            for coefficient in sensor_coefficients[sensor]:
                self._coefficients_xml[sensor][coefficient] = float(child.find(coefficient).text)


    def __get_hex_coefficients(self):
        
        """
        
        Retrieves calibration coefficients from SBE16/19 hex file (.hex) and stores them in self._coefficients_hex.
        
        """
        
        hex_file = open(self._filepath_hex, 'r')
        hex_header = hex_file.read()
        hex_file.close()

        hex_header = hex_header.split('*END*')[0]

        self._coefficients_hex = {
                                 'TemperatureSensor': {},
                                 'ConductivitySensor': {},
                                 'PressureSensor': {}
                                }

        self._coefficients_hex['TemperatureSensor']['A0'] = float(hex_header.split('<TA0>')[1].split('</TA0>')[0])
        self._coefficients_hex['TemperatureSensor']['A1'] = float(hex_header.split('<TA1>')[1].split('</TA1>')[0])
        self._coefficients_hex['TemperatureSensor']['A2'] = float(hex_header.split('<TA2>')[1].split('</TA2>')[0])
        self._coefficients_hex['TemperatureSensor']['A3'] = float(hex_header.split('<TA3>')[1].split('</TA3>')[0])

        self._coefficients_hex['ConductivitySensor']['G'] = float(hex_header.split('<G>')[1].split('</G>')[0])
        self._coefficients_hex['ConductivitySensor']['H'] = float(hex_header.split('<H>')[1].split('</H>')[0])
        self._coefficients_hex['ConductivitySensor']['I'] = float(hex_header.split('<I>')[1].split('</I>')[0])
        self._coefficients_hex['ConductivitySensor']['J'] = float(hex_header.split('<J>')[1].split('</J>')[0])
        self._coefficients_hex['ConductivitySensor']['CPcor'] = float(hex_header.split('<CPCOR>')[1].split('</CPCOR>')[0])
        self._coefficients_hex['ConductivitySensor']['CTcor'] = float(hex_header.split('<CTCOR>')[1].split('</CTCOR>')[0])

        self._coefficients_hex['PressureSensor']['PA0'] = float(hex_header.split('<PA0>')[1].split('</PA0>')[0])
        self._coefficients_hex['PressureSensor']['PA1'] = float(hex_header.split('<PA1>')[1].split('</PA1>')[0])
        self._coefficients_hex['PressureSensor']['PA2'] = float(hex_header.split('<PA2>')[1].split('</PA2>')[0])
        self._coefficients_hex['PressureSensor']['PTEMPA0'] = float(hex_header.split('<PTEMPA0>')[1].split('</PTEMPA0>')[0])
        self._coefficients_hex['PressureSensor']['PTEMPA1'] = float(hex_header.split('<PTEMPA1>')[1].split('</PTEMPA1>')[0])
        self._coefficients_hex['PressureSensor']['PTEMPA2'] = float(hex_header.split('<PTEMPA2>')[1].split('</PTEMPA2>')[0])
        self._coefficients_hex['PressureSensor']['PTCA0'] = float(hex_header.split('<PTCA0>')[1].split('</PTCA0>')[0])
        self._coefficients_hex['PressureSensor']['PTCA1'] = float(hex_header.split('<PTCA1>')[1].split('</PTCA1>')[0])
        self._coefficients_hex['PressureSensor']['PTCA2'] = float(hex_header.split('<PTCA2>')[1].split('</PTCA2>')[0])
        self._coefficients_hex['PressureSensor']['PTCB0'] = float(hex_header.split('<PTCB0>')[1].split('</PTCB0>')[0])
        self._coefficients_hex['PressureSensor']['PTCB1'] = float(hex_header.split('<PTCB1>')[1].split('</PTCB1>')[0])
        self._coefficients_hex['PressureSensor']['PTCB2'] = float(hex_header.split('<PTCB2>')[1].split('</PTCB2>')[0])
        
        
    def __compare_coefficients(self):

        """
        
        Compares calibration coefficients extracted from xml and hex file and warns user if they do not match.
        
        """
        
        for sensor in self._coefficients_hex:
            for coefficient in self._coefficients_hex[sensor]:
                rtol = abs(self._coefficients_xml[sensor][coefficient] * 0.001)
                if not np.isclose(self._coefficients_xml[sensor][coefficient], self._coefficients_hex[sensor][coefficient], rtol=rtol):
                    warnings.warn(f'*** WARNING *** : {sensor} calibration coefficient {coefficient} in xmlcon does not match corresponding value in hex file... Proceeding with value from xmlcon.')

    
    def __get_hex_samples(self):
        
        """
        
        Extracts hexadecimal measurements from SBE16/19 hex file. 
        
        """
        
        file_hex = open(self._filepath_hex, 'r')
        data_hex = file_hex.read()
        file_hex.close()
        
        self._hex_samples = data_hex.split('*END*')[1].strip().split('\n')
        
    def __parse_hexstr(self,x,start,end):
        """
        parse the hexadecimal string x from index start to end. assume it is 
        base 16 integer

        Parameters
        ----------
        x : string
            hex input string.
        start : int
            first character to parse
        end : int
            last_character to parse.

        Returns
        -------
        out : TYPE
            DESCRIPTION.

        """
        
        try:
            out = int(x[start:end], base=16)
        except:
            out = np.nan
        return out 
    
    
    def __hex2temperature(self):
        
        """
        
        Calculates temperature (degrees Celcius, C) from hexadecimal string.
        
        """
        
        A0 = self._coefficients_xml['TemperatureSensor']['A0']
        A1 = self._coefficients_xml['TemperatureSensor']['A1']
        A2 = self._coefficients_xml['TemperatureSensor']['A2']
        A3 = self._coefficients_xml['TemperatureSensor']['A3']
        
        temperature_decimal = np.array([int(x[:6], base=16) for x in self._hex_samples])
        MV = (temperature_decimal - 524288)/1.6e7
        R = ((MV*2.900e9) + 1.024e8) / (2.048e4 - (MV*2.0e5))
        temperature_degrees = 1/(A0 + A1*np.log(R) + A2*np.log(R)**2 + A3*np.log(R)**3) - 273.15
        
        self.data['Temperature (C)'] = temperature_degrees

    def __hex2time(self):
        global time_decimal
        time_decimal1 = np.array([self.__parse_hexstr(x,42,50) for x in self._hex_samples])
        
        #time_decimal = np.array([int(x[42:50],base = 16) for x in self._hex_samples])
        #print(time_decimal, time_decimal1)
        time = pd.to_datetime('01 Jan 2000')  + np.array([pd.Timedelta(t,'second') for t in time_decimal])
        
        self.time = time
        
    def __hex2conductivity(self):

        """
        
        Calculates conductivity (Simiens/meter, S/m) from hexadecimal string.
        
        """

        G = self._coefficients_xml['ConductivitySensor']['G']
        H = self._coefficients_xml['ConductivitySensor']['H']
        I = self._coefficients_xml['ConductivitySensor']['I']
        J = self._coefficients_xml['ConductivitySensor']['J']
        CPcor = self._coefficients_xml['ConductivitySensor']['CPcor']
        CTcor = self._coefficients_xml['ConductivitySensor']['CTcor']

        conductivity_decimal = np.array([int(x[6:12], base=16) for x in self._hex_samples])
        f = conductivity_decimal / 256 / 1000
        conductivity_spm = (G + H*(f**2)+ I*(f**3) + J*(f**4)) / (1 + CTcor*self.data['Temperature (C)'] + CPcor*self.data['Pressure (dbar)'])
        
        self.data['Conductivity (S/m)'] = conductivity_spm


    def __hex2pressure(self):

        """
        
        Calculates pressure (decibars, dbar) from hexadecimal string.
        
        """

        PA0 = self._coefficients_xml['PressureSensor']['PA0']
        PA1 = self._coefficients_xml['PressureSensor']['PA1']
        PA2 = self._coefficients_xml['PressureSensor']['PA2']
        PTEMPA0 = self._coefficients_xml['PressureSensor']['PTEMPA0']
        PTEMPA1 = self._coefficients_xml['PressureSensor']['PTEMPA1']
        PTEMPA2 = self._coefficients_xml['PressureSensor']['PTEMPA2']
        PTCA0 = self._coefficients_xml['PressureSensor']['PTCA0']
        PTCA1 = self._coefficients_xml['PressureSensor']['PTCA1']
        PTCA2 = self._coefficients_xml['PressureSensor']['PTCA2']
        PTCB0 = self._coefficients_xml['PressureSensor']['PTCB0']
        PTCB1 = self._coefficients_xml['PressureSensor']['PTCB1']
        PTCB2 = self._coefficients_xml['PressureSensor']['PTCB2']
        

    

        pressure_decimal = np.array([self.__parse_hexstr(x,12,18) for x in self._hex_samples])
        ptcv_decimal = np.array([self.__parse_hexstr(x,18,22) for x in self._hex_samples])
        
        y = ptcv_decimal / 13107
        t = PTEMPA0 + PTEMPA1*(y) + PTEMPA2*(y**2)
        x = pressure_decimal - PTCA0 - PTCA1*t - PTCA2*(t**2)
        n = (x*PTCB0) / (PTCB0 + PTCB1*t + PTCB2*(t**2))
        
        pressure_dbar = PA0 + PA1*n + PA2*(n**2)
        pressure_dbar = (pressure_dbar - 14.7) * 0.689476
        
        self.data['Pressure (dbar)'] = pressure_dbar


    def __hex2turbidity(self):
        
        """
        
        Calculates turbidity (Nephelometric turbidity, NTU) from hexadecimal string measured 
        by WET Labs ECO NTU turbidity sensor.
        
        """
        
        scale_factor = self._coefficients_xml['TurbidityMeter']['ScaleFactor']
        dark_voltage = self._coefficients_xml['TurbidityMeter']['DarkVoltage']
        #print(self._xmlcon['Name'], self._sensor_index['TurbidityMeter'])
        
        
        start_idx = 4*(self._sensor_index['TurbidityMeter'] - 3) + 22
        end_idx = start_idx + 4
        #print(start_idx, end_idx)
        
        turbidity_decimal = np.array([self.__parse_hexstr(x,start_idx,end_idx) for x in self._hex_samples])
        # if self._xmlcon['Name'] == 'SBE 16plus V2 Seacat CTD':
            
        #     turbidity_decimal = np.array([self.__parse_hexstr(x,26,30) for x in self._hex_samples])#np.array([int(x[26:30], base=16) for x in self._hex_samples])
            
        # elif self._xmlcon['Name'] == 'SBE 19plus V2 Seacat CTD':
            
        #     turbidity_decimal = np.array([self.__parse_hexstr(x,30,34) for x in self._hex_samples])#np.array([int(x[30:34], base=16) for x in self._hex_samples])
        
        turbidity_volt = turbidity_decimal / 13107
        turbidity_ntu = scale_factor * (turbidity_volt - dark_voltage)
        
        self.data['Turbidity (NTU)'] = turbidity_ntu
        
        
    def __hex2transmission(self):
        
        """
        
        Calculates light transmission (%) and beam attenuation coefficients from hexadecimal string measured 
        by WET Labs C-Star transmissometer.
        
        """
        
        M = self._coefficients_xml['WET_LabsCStar']['M']
        B = self._coefficients_xml['WET_LabsCStar']['B']
        z = self._coefficients_xml['WET_LabsCStar']['PathLength']
        
        start_idx = 4*(self._sensor_index['WET_LabsCStar'] - 3) + 22
        end_idx = start_idx + 4
        #print(start_idx, end_idx)
        
        transmission_decimal = np.array([self.__parse_hexstr(x,start_idx,end_idx) for x in self._hex_samples])
        #transmission_decimal = np.array([self.__parse_hexstr(x,22,26) for x in self._hex_samples])#np.array([int(x[22:26], base=16) for x in self._hex_samples])
        transmission_volt = transmission_decimal / 13107
        
        beam_transmission = (M*transmission_volt) + B
        beam_attenuation = -(1/z)*np.log(beam_transmission/100)
        
        self.data['Light Transmission (%)'] = beam_transmission
        self.data['Beam Attenuation Coefficient'] = beam_attenuation
    
    
    def __hex2oxygen(self):
        
        """
        
        Calculates oxygen concentration (milliliters/Liter, mL/L) from hexadecimal string.
        For details on calculation of salinity correction factor (Scorr) see Appendix I of:
        https://www.seabird.com/asset-get.download.jsa?id=54627862513
        
        """
        start_idx = 4*(self._sensor_index['OxygenSensor'] - 3) + 22
        #end_idx = start_idx + 4
        #print(start_idx, end_idx)
        # SBE 63 Dissolved Oxygen Sensor calculations
        if self._xmlcon['Name'] == 'SBE 16plus V2 Seacat CTD':
            end_idx = start_idx + 6
            A0 = self._coefficients_xml['OxygenSensor']['A0']
            A1 = self._coefficients_xml['OxygenSensor']['A1']
            A2 = self._coefficients_xml['OxygenSensor']['A2']
            B0 = self._coefficients_xml['OxygenSensor']['B0']
            B1 = self._coefficients_xml['OxygenSensor']['B1']
            C0 = self._coefficients_xml['OxygenSensor']['C0']
            C1 = self._coefficients_xml['OxygenSensor']['C1']
            C2 = self._coefficients_xml['OxygenSensor']['C2']
            E = self._coefficients_xml['OxygenSensor']['pcor']
            
            T = self.data['Temperature (C)']
            P = self.data['Pressure (dbar)']
            S = self.data['Salinity (PSU)']  
            
            # Coefficients for salinity correction factor (Scorr)
            solB0 = -6.24523e-3; solB1 = -7.37614e-3; solB2 = -1.03410e-2; solB3 = -8.17083e-3
            solC0 = -4.88682e-7
            
            Ts = np.log((298.15-T)/(273.15+T))
            
            Scorr = np.exp(S*(solB0+(solB1*Ts)+(solB2*(Ts**2))+(solB3*(Ts**3))) + (solC0*(S**2)))
            
            

            oxygen_decimal = np.array([self.__parse_hexstr(x,start_idx,end_idx) for x in self._hex_samples])#np.array([int(x[30:36], base=16) for x in self._hex_samples])
            
            
            #oxygen_decimal = np.array([self.__parse_hexstr(x,30,36) for x in self._hex_samples])#np.array([int(x[30:36], base=16) for x in self._hex_samples])
            
            V = ((oxygen_decimal/100000)-10) / 39.457071 

            eqpt1 = ((A0+(A1*T)+(A2*(V**2))) / (B0+(B1*V))) - 1
            eqpt2 = C0+(C1*T)+(C2*(T**2))
            eqpt3 = np.exp(E*P/(T+273.15))
            
            oxygen_conc = (eqpt1/eqpt2)*eqpt3*Scorr
            
            self.data['Oxygen (ml/l)'] = oxygen_conc
        
        # SBE 43 Dissolved Oxygen Sensor calculations
        elif self._xmlcon['Name'] == 'SBE 19plus V2 Seacat CTD':
            end_idx = start_idx + 4
            A = self._coefficients_xml['OxygenSensor']['A']
            B = self._coefficients_xml['OxygenSensor']['B']
            C = self._coefficients_xml['OxygenSensor']['C']
            #D0 = self._coefficients_xml['OxygenSensor']['D0']
            #D1 = self._coefficients_xml['OxygenSensor']['D1']
            #D2 = self._coefficients_xml['OxygenSensor']['D2']
            E = self._coefficients_xml['OxygenSensor']['E']
            #Tau20 = self._coefficients_xml['OxygenSensor']['Tau20']
            #H1 = self._coefficients_xml['OxygenSensor']['H1']
            #H2 = self._coefficients_xml['OxygenSensor']['H2']
            #H3 = self._coefficients_xml['OxygenSensor']['H3']
            Soc = self._coefficients_xml['OxygenSensor']['Soc']
            Voffset = self._coefficients_xml['OxygenSensor']['offset']
            
            T = self.data['Temperature (C)']
            P = self.data['Pressure (dbar)']
            
            
            oxygen_decimal = np.array([self.__parse_hexstr(x,start_idx,end_idx) for x in self._hex_samples])
            #oxygen_decimal = np.array([self.__parse_hexstr(x,26,30) for x in self._hex_samples])#np.array([int(x[26:30], base=16) for x in self._hex_samples])
            
            V = oxygen_decimal / 13107
            
            eqpt1 = Soc*(V+Voffset)*(1+(A*T)+(B*(T**2))+(C*(T**3)))
            eqpt2 = np.exp(E*P/(T+273.15))
            oxysat = self.__oxygen_saturation()
            
            oxygen_conc = eqpt1*eqpt2*oxysat
            
            self.data['Oxygen (ml/l)'] = oxygen_conc
    
    
    def __conductivity2salinity(self):
        
        """
        
        Calculates salinity (PSU) from conductivity following the standards of the Practical Salinity Scale - 1978 (PSS-78)
        For details on calculation see:
        https://salinometry.com/pss-78/
        https://www.seabird.com/asset-get.download.jsa?id=54627861526
        
        """
        
        a = [0.008, -0.1692, 25.3851, 14.0941, -7.0261, 2.7081]
        b = [0.0005, -0.0056, -0.0066, -0.0375, 0.0636, -0.0144]
        k = 0.0162
        
        A1 = 2.07e-5; A2 = -6.37e-10; A3 = 3.989e-15
        B1 = 3.426e-2; B2 = 4.464e-4; B3 = 4.215e-1; B4 = -3.107e-3
        C0 = 6.766097e-1; C1 = 2.00564e-2; C2 = 1.104259e-4; C3 = -6.9698e-7; C4 = 1.0031e-9
        
        T = self.data['Temperature (C)'] * 1.00024      # Temperature IPTS-68, C
        P = self.data['Pressure (dbar)']                # Pressure, dbars
        R = self.data['Conductivity (S/m)'] / 4.2914    # Conductivity ratio (C of water at 35 PSU and 15C @ sea-level = 4.2914 S/m)
        
        Rp = 1 + (((A1*P)+(A2*(P**2))+(A3*(P**3)))/(1+(B1*T)+(B2*(T**2))+(B3*R)+(B4*T*R)))
        rT = C0 + (C1*T) + (C2*(T**2)) + (C3*(T**3)) + (C4*(T**4))
        RT = R/(Rp*rT)
        
        S1 = 0; S2 = 0
        for i in range(6):
            S1 += a[i] * np.power(RT,(i/2))
            S2 += b[i] * np.power(RT,(i/2))
        
        S = S1 + (((T-15)/(1+(k*(T-15))))*S2)
        
        self.data['Salinity (PSU)'] = S
       
        
    def __oxygen_saturation(self):
        
        """
        
        Calculates the oxygen saturation limit (mL/L) as a function of temperature and salinity 
        For details see Appendix A of:
        http://www.argodatamgt.org/content/download/26535/181243/file/SBE43_ApplicationNote64_RevJun2013.pdf
        
        Returns:
            oxysat (np.array): oxygen saturation limit (mL/L)
        
        """
        
        A0 = 2.00907; A1 = 3.22014; A2 = 4.0501; A3 = 4.94457; A4 = -0.256847; A5 = 3.88767
        B0 = -0.00624523; B1 = -0.00737614; B2 = -0.010341; B3 = -0.00817083
        C0 = -4.88682e-7
        
        T = self.data['Temperature (C)']
        S = self.data['Salinity (PSU)']
        
        Ts = np.log((298.15-T)/(273.15+T))
        
        eqpt1 = A0+(A1*Ts)+(A2*(Ts**2))+(A3*(Ts**3))+(A4*(Ts**4))+(A5*(Ts**5))
        eqpt2 = (S*(B0+(B1*Ts)+(B2*(Ts**2))+(B3*(Ts**3)))) + (C0*(S**2))
        
        oxysat = np.exp(eqpt1+eqpt2)
        
        return oxysat
        
            

        
        
        return df
        
#%%




#%%  

class DataSet(seabird_ctd):
    """
  A derived class from seabird_ctd to enhance data handling and analytics specific 
  to sea-bird conductivity, temperature, and depth (CTD) instruments.

  This class integrates additional functionalities for plotting, geometric calculations,
  data processing, and masking operations on top of the base functionalities provided
  by the seabird_ctd class.

  Attributes:
      plot (__Plotting): An instance of the nested __Plotting class for handling data visualization.
      name (str): The filename extracted from the filepath_hex, used as an identifier.
      geometry (__Geometry): An instance of the nested __Geometry class for managing sensor geometry.
      mask (__Masking): An instance of the nested __Masking class for applying masks to the data.
      processing (__Processing): An instance of the nested __Processing class for additional data processing tasks.

  Parameters:
      filepath_hex (str): Path to the hexadecimal data file from the CTD instrument.
      filepath_xml (str): Path to the XML configuration file associated with the CTD data.
      **kwargs: Arbitrary keyword arguments that can be passed to the seabird_ctd constructor.

  Methods:
      get_timeseries_data(field_name='Turbidity (NTU)', mask=True): Retrieves timeseries data for a specified field.

     """   
    def __init__(self,filepath_hex, filepath_xml,**kwargs):
        super().__init__(filepath_hex, filepath_xml)
        self.plot = self.__Plotting(self)
        self.name = kwargs.pop('name',filepath_hex.split(os.sep)[-1])
        self.geometry = self.__Geometry(self)
        self.mask = self.__Masking(self)
        self.processing = self.__Processing(self)
        self.timezone_correction = 0

        
    def get_timeseries_data(self, field_name='Turbidity (NTU)', mask=True):
        """
        Retrieve timeseries data for a given field name.
    
        Parameters:
        - field_name (str): Name of the field for which timeseries data is retrieved (default: 'Turbidity (NTU)').
        - mask (bool): If True, any defined masks will be applied to the data (default: True).
    
        Returns:
        - t (numpy.ndarray): DateTime array representing the time values.
        - x (numpy.ndarray): Timeseries data array for the specified field_name.
        """
        t = self.data['DateTime'].copy() + pd.to_timedelta(f'{self.timezone_correction}h') #apply timezone correction
        x = self.data[field_name].copy()
    
        if mask: # apply masks 
            x = self.mask.apply_masks(x)
    
        return t, x
    
    def to_csv(self):
        
        df = self.to_pandas_df()
        
        df['DateTime'] = df['DateTime'] + pd.to_timedelta(f'{self.timezone_correction}h') #apply timezone correction
        
        filename_hex = os.path.basename(self._filepath_hex)
        filename_csv = os.path.splitext(filename_hex)[0] + '.csv'
        
        df.to_csv(filename_csv)
        
    def to_pandas_df(self, mask=True):
        df = pd.DataFrame.from_dict(self.data)
        df['DateTime'] = df['DateTime'] + pd.to_timedelta(f'{self.timezone_correction}h') #apply timezone correction
        df.set_index('DateTime',drop = True,inplace = True)
        
        return df
        
    class __Processing:
        """
        Manage data processing functionality for CTD data.
    
        Parameters
        ----------
        ctd : CTD
            CTD (Conductivity, Temperature, Depth) object.
    
        Methods
        -------
        calculate_depth(environment='SALTWATER')
            Calculate depth from pressure following SBE APPLICATION NOTE NO. 69.
    
        Attributes
        ----------
        _ctd : CTD
            CTD object containing data for processing.
        """
    
        def __init__(self, ctd):
            self._ctd = ctd
            
            
        def calculate_SSC(self, A=1):
            """
            Calculates Suspended Solids Concentration (SSC) from turbidity data.
        
            Parameters:
            - self (object): Instance of the class containing CTD data.
            - A (float): Conversion factor from Turbidity (NTU) to SSC.
        
            Returns:
            None. Updates the CTD data with the calculated SSC values.
            """
            # convert from turbidity (NTU) to suspended solids concentration assuming SSC = A * NTU
            self._ctd.data['SSC (mg/L)'] = A * self._ctd.data['Turbidity (NTU)']
            
    
        def calculate_depth(self, environment='SALTWATER'):
            """
            Calculate depth from pressure following SBE APPLICATION NOTE NO. 69.
    
            Sea-Bird uses the formula in UNESCO Technical Papers in Marine Science
            No. 44. This is an empirical formula that takes compressibility (that is,
            density) into account. An ocean water column at 0 °C (t = 0) and 35 PSU
            (s = 35) is assumed.
            
            If latitude has not been calculated in the position
            dataset, then both latitude and longitude will be calculated 
            automatically
    
            Parameters
            ----------
            environment : str, optional
                Specify whether the CTD is in "SALTWATER" or "FRESHWATER".
                The default is 'SALTWATER'.
    
            Returns
            -------
            None
            """
            
            

    
            p = self._ctd.data['Pressure (dbar)'] # Pressure
            latitude = self._ctd.geometry.pose.pose['latitude'].to_numpy()  # 
            x = np.sin(latitude / 57.29578) ** 2
            g = (9.780318) * (1.0 + ((5.2788e-3) + (2.36e-5) * x) * x) + (1.092e-6) * p  # Gravity variation with latitude and pressure
            d = (((( (-1.82e-15) * p + 2.279e-10) * p - 2.2512e-5) * p + 9.72659) * p) / g
            self._ctd.data['Depth (m)'] = d  # Update the data dictionary
                
            
        def calculate_density(self):
            """
            Calculate the density of seawater using UNESCO 1983 polynomial equations of state.
            
            Validated agaist this online calucator https://www.phys.ocean.dal.ca/~kelley/seawater/density.html
            Parameters:
            - s : numpy.ndarray
                Salinity in Practical Salinity Units (PSU).
            - t : numpy.ndarray
                Temperature in degrees Celsius.
            - p : numpy.ndarray
                Pressure in decibars.
            
            Returns:
            - numpy.ndarray
                Density of seawater.
            """
            t = np.array(self._ctd.data['Temperature (C)'])
            p = np.array(self._ctd.data['Pressure (dbar)'])
            s = np.array(self._ctd.data['Salinity (PSU)'])
            
            A0 = 999.842594
            A1 = 6.793952e-2
            A2 = -9.095290e-3
            A3 = 1.001685e-4
            A4 = -1.120083e-6
            A5 = 6.536332e-9
            
            B0 = 8.24493e-1
            B1 = -4.0899e-3
            B2 = 7.6438e-5
            B3 = -8.2467e-7
            B4 = 5.3875e-9
            
            C0 = -5.72466e-3
            C1 = 1.0227e-4
            C2 = -1.6546e-6
            
            D0 = 4.8314e-4
            
            
            
            FQ0 = 54.6746
            FQ1 = -0.603459
            FQ2 = 1.09987e-2
            FQ3 = -6.1670e-5
            
            G0 = 7.944e-2
            G1 = 1.6483e-2
            G2 = -5.3009e-4
            
            i0 = 2.2838e-3
            i1 = -1.0981e-5
            i2 = -1.6078e-6
            
            J0 = 1.91075e-4
            
            M0 = -9.9348e-7
            M1 = 2.0816e-8
            M2 = 9.1697e-10
            
            E0 = 19652.21
            E1 = 148.4206 #c
            E2 = -2.327105
            E3 = 1.360477e-2
            E4 = -5.155288e-5
            
            H0 = 3.239908
            H1 = 1.43713e-3
            H2 = 1.16092e-4
            H3 = -5.77905e-7
            
            K0 = 8.50935e-5
            K1 = -6.12293e-6
            K2 = 5.2787e-8
            
            t2 = t * t
            t3 = t * t2
            t4 = t * t3
            t5 = t * t4
            
            #s = np.maximum(s, 0.000001)  # Avoid division by zero
            
            s32 = np.power(s, 1.5)
            
            p = p/10.0  # convert decibars to bars
            
            sigma = A0 + A1*t + A2*t2 + A3*t3 + A4*t4 + A5*t5+ (B0 + B1*t + B2*t2 + B3*t3 + B4*t4)*s+ (C0 + C1*t + C2*t2)*s32 + D0*s*s
            
            kw = E0 + E1*t + E2*t2 + E3*t3 + E4*t4
            aw = H0 + H1*t + H2*t2 + H3*t3
            bw = K0 + K1*t + K2*t2
            
            k = kw + (FQ0 + FQ1*t + FQ2*t2 + FQ3*t3)*s + (G0 + G1*t + G2*t2)*s32+ (aw + (i0 + i1*t + i2*t2)*s + (J0*s32))*p + (bw + (M0 + M1*t + M2*t2)*s)*p*p
            
            val = 1.0-(p/k)
            
            sigma = (sigma/val) #- 1000
                # non_zero_mask = val != 0.0
                # sigma[non_zero_mask] = sigma[non_zero_mask] / val[non_zero_mask] - 1000.0
            self._ctd.data['Density (kg/m3)'] = sigma


    class __Geometry:
        """
        Manage the geometry (position, orientation, etc...) of an instrument.
    
        Parameters
        ----------
        ctd : CTD
            CTD (Conductivity, Temperature, Depth) object.
    
        Attributes
        ----------
        relative_position : numpy array
            Position of the sensor relative to the position measurement for the instrument.
        Pose : Pose
            Pose object containing the instrument's position and orientation at all timestamps.
    
        Methods
        -------
        set_Pose(Pose)
            Define the instrument pose (position and orientation) with a Pose object.
    
        set_sensor_relative_geometry(rotation=0, offset=(0, 0, 0))
            Calculate the x, y, z position of the sensor in the coordinate reference frame of the instrument.
    
        get_absolute_sensor_position()
            Calculate the absolute position of the sensor, taking into account the position and orientation of the platform.
        """
    
        def __init__(self, ctd):
            self._ctd = ctd
            self.relative_position = np.zeros(3, dtype=float)
            self.pose = None
    
        def set_pose(self, Pose, update_depth = False,**kwargs):
            """
            Define the instrument pose (position and orientation) with a 
            Pose object. Makes a cope of the pose object and stores it under 
            ctd.geometry.Pose
            
            Parameters
            ----------
            Pose : Pose
                Pose object containing position and orientation information.
            update_depth : bool, optional 
                If True, the calculated depth (from ctd pressure) will be used
                as the pose z variable data. 
    
            Returns
            -------
            None
            """
            
            
            
            
            
            # copy input object and Resample the input timeseries to the CTD measurement timesteps.
            self.pose = copy.deepcopy(Pose)
            
            # ## infer the best tolerance 
            # t = self._ctd.data['DateTime'].to_numpy()
            # dt = np.diff(t,prepend = t[0],).astype(float)/1e9 # time delta in seconds
            # tolerance = 1.01*np.percentile(dt,99) # mask to identify data gaps 
            # print(tolerance)
            
            
            self.pose.resample_to(self._ctd.data['DateTime'],kwargs)
    
            if not self.pose.pose.get('latitude'):
                self._ctd.geometry.pose.add_lat_lon()

            
            # update pose z variable with calculated depth from ctd data
            if update_depth:
                # calculate latitude and longitude if needed
                # if not self.pose.pose.get('latitude'):
                #     self._ctd.geometry.pose.add_lat_lon()
                self._ctd.geometry.pose.add_lat_lon()
                # calculate depth (based on pressure and latitude)
                self._ctd.processing.calculate_depth()
                
                # resample input timeser
                self._ctd.geometry.pose.resample_from(pd.Series(-self._ctd.data['Depth (m)'], index =  self._ctd.data['DateTime']),field_name = 'z', z_convention = 'reverse')
                
                

                
            
            
    
        def set_sensor_relative_geometry(self, rotation=0, offset=(0, 0, 0)):
            """
            Calculate the x, y, z position of the sensor in the coordinate reference frame of the instrument.
    
            Parameters
            ----------
            rotation : float, optional
                Rotation angle (in degrees) of the sensor.
                The default is 0.
    
            offset : numpy array, optional
                x, y, z offsets for the instrument. This offset should be the relative distance between the
                position measurement (e.g., from the vessel GPS) and the center of the sensor.
                The default is (0, 0, 0).
    
            Returns
            -------
            None
            """
            self.relative_position = np.array(offset, dtype=float)  # Position of the sensor relative to the instrument.
    
        def get_absolute_sensor_position(self):
            """
            Calculate the absolute position of the sensor, taking into account the position and orientation
            of the platform (e.g., ROV), and the relative position of the sensor on the platform.
    
            Returns
            -------
            absolute_position: numpy array
                numpy array containing the (x, y, z) position of the sensor for every timestep.
            """
            # Get position of beam midpoints for every ensemble.
            X = np.empty((3, self._ctd.n_samples))
            pose = self.pose.pose
    
            for i in range(self._ctd.n_samples):
                # Build rotation matrix.
                yaw = pose['heading'].to_numpy()[i]
                pitch = pose['pitch'].to_numpy()[i]
                roll = pose['roll'].to_numpy()[i]
                R = np.dot(ptools.gen_rot_x(roll), ptools.gen_rot_z(yaw).dot(ptools.gen_rot_y(pitch)))  # 3-d rotation operator
                Xp = np.add(pose[['x', 'y', 'z']].to_numpy().T[:, i], self.relative_position.T.dot(R)).T  # Apply rotation to sensor relative orientation vector
                X[:, i] = Xp  # Update position for the current sample
    
            # Apply masks (for x, y, z coordinates).
            X[0] = self._ctd.mask.apply_masks(X[0])
            X[1] = self._ctd.mask.apply_masks(X[1])
            X[2] = self._ctd.mask.apply_masks(X[2])
    
            return X
            
    
    class __Plotting:
        """
        Manage plotting functionality for CTD data.
    
        Parameters
        ----------
        ctd : CTD
            CTD (Conductivity, Temperature, Depth) object.
    
        Methods
        -------
        timeseries(field_name='Turbidity (NTU)')
            Create a timeseries plot of CTD data.
    
        Attributes
        ----------
        _ctd : CTD
            CTD object containing data for plotting.
        """
    
        def __init__(self, ctd):
            self._ctd = ctd
    
        def timeseries(self, field_name='Turbidity (NTU)',mask = True):
            """
            Create a timeseries plot of CTD data.
    
            Parameters
            ----------
            field_name : str, optional
                Name of the field to plot. Any key in the CTD data dictionary or 'all' to plot all fields in subplots.
                The default is 'Turbidity (NTU)'. 
            mask: bool, optional
                if True, masks will be applied to the data (default). If false, raw data will be plotted. 
            Returns
            -------
            fig : matplotlib.figure.Figure
                Matplotlib figure object.
    
            ax : matplotlib.axes._axes.Axes
                Matplotlib axes object.
            """
            
            #if field_name == 'all':
            
            
            
            fig, ax = plt.subplots(figsize=(10, 4))
            x,y = self._ctd.get_timeseries_data(field_name=field_name,mask = mask)
            ax.plot(x, y, lw=1)
            ax.set_ylabel(field_name)
            ax.set_xlabel('Time')
    
            xlims = mdates.date2num(x)
    
            ax.xaxis.set_major_formatter(mdates.DateFormatter('%H:%M %d%b%y '))
            ax.grid(alpha=0.1)
            ax.set_xticklabels(ax.get_xticklabels(), rotation=0, ha='center')
            ax.xaxis.tick_bottom()
            ax.xaxis.set_label_position('bottom')
            
            title = self._ctd._filepath_hex.split(os.sep)[-1] + '\n' + self._ctd._filepath_xml.split(os.sep)[-1]
            ax.set_title(title, fontsize = 7)
    
            return fig, ax
                
            
    class __Masking:
        """
        Manage masking functionality for CTD data.
    
        Parameters
        ----------
        ctd : CTD
            CTD (Conductivity, Temperature, Depth) object.
    
        Attributes
        ----------
        _ctd : CTD
            CTD object containing data for masking.
        masks : dict
            Dictionary to hold masks for ensemble data.
        mask_status : dict
            Dictionary to hold activation status of defined masks.
    
        Methods
        -------
        define_mask(mask, name, set_active=False)
            Define a boolean mask that can be applied to timeseries data.
            
        set_mask_status(status, name='all')
            Change the status of a mask, or multiple masks.
            
        apply_masks(X)
            Apply all active masks to the ensemble array X.
        """
    
        def __init__(self, ctd):
            self._ctd = ctd
            self.masks = {}  # Dictionary to hold masks for ensemble data
            self.mask_status = {}  # Dictionary to hold activation status of defined masks
    
        def define_mask(self, mask, name, set_active=False):
            """
            Define a boolean mask that can be applied to timeseries data.
    
            Parameters
            ----------
            mask : numpy array
                Numpy array with dimensions (n_samples).
            name : str
                Name for the mask.
            set_active : bool, optional
                If True, the defined mask will be activated.
                Default is False.
    
            Returns
            -------
            None
            """
            # Check that mask has correct dimensions
            valid_dims = (self._ctd.n_samples)
            if not np.shape(mask) == valid_dims:
                ValueError(f'Mask must have dimensions of {valid_dims}')
    
            self.masks[name] = mask
    
            # Initialize the mask status
            if set_active:
                self.mask_status[name] = True
            else:
                self.mask_status[name] = False
    
        def set_mask_status(self, status, name='all'):
            """
            Change the status of a mask, or multiple masks.
            The ensemble data corresponding to FALSE entries in the mask will be converted to np.nan.
    
            Parameters
            ----------
            name : str or list of str, optional
                Mask name, or list of mask names. If 'all', all defined masks will be activated.
                The default is 'all'.
            status : bool
                New status for the mask(s).
    
            Raises
            ------
            ValueError
                Invalid mask name.
    
            Returns
            -------
            None
            """
            if name == 'all':
                for key in self.masks.keys():
                    self.mask_status[key] = status
            elif type(name) == list:
                for n in name:
                    # Check if mask exists
                    if n in self.masks.keys():
                        self.mask_status[n] = status
                    else:
                        raise ValueError(f'{n} is not a defined mask.')
            elif type(name) == str:
                if name in self.masks.keys():
                    self.mask_status[name] = status
                else:
                    raise ValueError(f'{name} is not a defined mask.')
    
        def apply_masks(self, X):
            """
            Apply all of the active masks to the ensemble array X.
            Masked values will be converted to np.nan.
    
            Parameters
            ----------
            X : numpy array
                Array of ensemble data with dimensions (n_bins, n_ensembles).
    
            Returns
            -------
            X : numpy array
                Masked array of ensemble data with dimensions (n_bins, n_ensembles).
            """
            if self.masks:
                for mask_name in self.masks.keys():
                    if self.mask_status[mask_name]:
                        mask = self.masks[mask_name]
                        np.putmask(X, ~mask, values=np.nan)
    
            return X
    

#%%


# hex_fpath = r'P:\41806287\41806287 NORI-D Data\Data\ROV\Hidden Gem\CTD\Raw\ROVDive20220921\DHIDive4_220921_1.hex'
# xmlcon_fpath = r'P:\41806287\41806287 NORI-D Data\Data\ROV\Hidden Gem\CTD\Raw\ROVDive20220921\DHIDIVE4_220921_1.xmlcon'
# ctd1 = seabird_ctd(filepath_hex = hex_fpath, filepath_xml = xmlcon_fpath)
# print(ctd1._sensor_index)

# hex_fpath = r'P:\41806287\41806287 NORI-D Data\Data\ROV\Island Pride\CTD\Raw\ROV_CTD_09102022\SBE19plus_01907947_2022_10_09_0005.hex'
# xmlcon_fpath = r'P:\41806287\41806287 NORI-D Data\Data\ROV\Island Pride\CTD\Config\SBE19-7947_DO-1900_CStar-1917_NTU-5874_xmiss_DO_swap.xmlcon'
# ctd2 = seabird_ctd(filepath_hex = hex_fpath, filepath_xml = xmlcon_fpath)
# print(ctd2._sensor_index)
#ctd = DataSet(filepath_hex = hex_fpath, filepath_xml = xmlcon_fpath)
#%
#%%
# tree = ET.parse(ctd._filepath_xml)
# root = tree.getroot()



# sensors = {}

# for sensor in root.findall('Instrument/SensorArray/'):
#     for s1 in sensor:
#         sensors[s1.tag] = {}
        
        
#         #sensors['Sen'] = 
#         #{'SensorID': '38'}
#         print(s1.attrib)
#         for s2 in s1:
#             sensors[s1.tag][s2.tag] = {}
#             for s3 in s2:
#                 print(s1.tag,s2.tag,s3.tag)
#                 sensors[s1.tag][s2.tag][s3.tag] = s3.text #pd.to_numeric(s3.text)
        
        
        # print(sensor.attrib)
        # print(child)

#%% read the xmlcon to get sensor order 





#%%

# xml_filepath = r'\\USDEN1-STOR.DHI.DK\\Projects\\41806287\\41806287 NORI-D Data\\Data\\ROV\\Island Pride HD14\\CTD\\Raw\\ROV_CTD_07102022\\Configuration\\SBE19-7947_DO-1900_CStar-1917_NTU-5874_xmiss_DO_swap.xmlcon'
# hex_filepath = r'\\USDEN1-STOR.DHI.DK\\Projects\\41806287\\41806287 NORI-D Data\\Data\\ROV\\Island Pride HD14\\CTD\\Raw\\ROV_CTD_27102022\\SBE19plus_01907947_2022_10_27_0026.hex'
# ctd = DataSet(hex_filepath,xml_filepath)

# x,t = ctd.get_timeseries_data()

# mask = (x>0.2)
# ctd.mask.define_mask(mask = mask, name = 'turbidity',set_active = True)


# ctd.plot.timeseries_plot()


# def main():     
    
#     scriptpath = os.path.abspath('.')
    
#     relpath_xml = 'ROV CTD/Config/SBE19-7947_DO-1900_CStar-1917_NTU-5874_xmiss_DO_swap.xmlcon'
#     #relpath_xml = 'FBCT1 CTD/Config/16-50406.xmlcon'
#     #relpath_xml = 'DeckSamplingSBE/DECKSMPL_RETURN03.XMLCON'

#     relpath_hex = 'ROV CTD/Raw/SBE19plus_01907947_2022_10_19_0001.hex'
#     #relpath_hex = 'FBCT1 CTD/Raw/SBE16plus_01650406_2022_10_04.hex'
#     #relpath_hex = 'DeckSamplingSBE/decksmpl_return03.hex'
    
#     filepath_xml = os.path.join(scriptpath,relpath_xml)
#     filepath_hex = os.path.join(scriptpath,relpath_hex)
    
#     test = SeabirdCTD(filepath_hex, filepath_xml)
    
#     #print(test.number_samples)
#     #print(test.data)
#     #print(test.coefficients_xml)
#     #test.dataframe2csv()
    
# ###############################################################################

# if __name__ == '__main__':
#     main()
    
###############################################################################
